# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/
"""Some widget construction to check if a sample moved"""

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "19/03/2018"


import logging
from collections import OrderedDict
from typing import Union

import h5py
from silx.gui import qt
from tomoscan.factory import Factory
from tomoscan.identifier import VolumeIdentifier

from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.gui.qfolderdialog import QDataDialog
from tomwer.gui.stackplot import _QImageFileStackPlot

_logger = logging.getLogger(__name__)


class _ImageStack(qt.QMainWindow):
    """
    This widget will make copy or virtual link to all received *slice* files
    in order to group them all in one place and be able to browse those
    (using the image stack of view in orange or a third software as silx view)

    Options are:
       - copy files or create sym link (set to sym link)
       - overwrite if existing (set to False)

    Behavior:
        When the process receives a new data path ([scanPath]/[scan]) and if
        no output folder has been defined manually them it will try to create
        the folder [scanPath]/slices if not existing in order to redirect
        the slices files.
        If fails will ask for a directory.
        If the output folder is already existing then move directly to the
        copy.
    """

    def __init__(self, parent=None, backend=None):
        qt.QMainWindow.__init__(self, parent)
        self.setWindowFlags(qt.Qt.Widget)
        self._scans = set()
        self._viewer = _QImageFileStackPlot(parent=self, backend=backend)
        self._viewer.addFolderName(True)
        self.setCentralWidget(self._viewer)

        self._viewer.getControlWidget().hide()

        self._dockWidgetMenu = qt.QDockWidget(parent=self)
        self._dockWidgetMenu.layout().setContentsMargins(0, 0, 0, 0)
        self._dockWidgetMenu.setFeatures(qt.QDockWidget.DockWidgetMovable)
        self._dockWidgetMenu.setWidget(self._viewer.getControlWidget())
        self.addDockWidget(qt.Qt.BottomDockWidgetArea, self._dockWidgetMenu)

        toolbar = qt.QToolBar()
        self.addToolBar(qt.Qt.RightToolBarArea, toolbar)

        self._clearAction = _ClearAction(parent=toolbar)
        toolbar.addAction(self._clearAction)
        self._clearAction.triggered.connect(self.clear)

        self._addAction = _AddAction(parent=toolbar)
        toolbar.addAction(self._addAction)
        self._addAction.triggered.connect(self._addNewFolder)
        # expose API
        self.setLoadingMode = self._viewer.setLoadingMode
        self.setForceSync = self._viewer.setForceSync

        self.setLoadingMode("lazy loading")

    def addLeafScan(self, scanID: Union[str, TomwerScanBase]):
        self._scans.add(scanID)
        self._update()

    def _addNewFolder(self):
        dialog = QDataDialog(self, multiSelection=True)

        if not dialog.exec_():
            dialog.close()
            return

        for folder in dialog.files_selected():
            self.addLeafScan(folder)

    def removeLeafScan(self, scanID):
        self._scans.remove(str(scanID))
        self._update()

    def extractImages(self):
        """
        Parse all self._scans and find the images to be displayed on the widget

        :return: images to display for each scan
        :rtype: dict
        """
        raise NotImplementedError("Base class")

    def _update(self):
        all_images = self.extractImages()
        all_images = OrderedDict(sorted(all_images.items()))
        ordered_imgs = []
        for scan in all_images:
            try:
                oScanSlices = OrderedDict(sorted(all_images[scan].items()))
                for scan_slice in oScanSlices:
                    ordered_imgs.append(all_images[scan][scan_slice])
            except TypeError:
                # in case algorithm could extract recons index
                ordered_imgs = ordered_imgs + list(all_images[scan].values())
        self._viewer.setImages(ordered_imgs)

    def clear(self):
        self._scans = set()
        self._viewer.clear()


class _ClearAction(qt.QAction):
    def __init__(self, parent):
        style = qt.QApplication.instance().style()
        icon = style.standardIcon(qt.QStyle.SP_DialogResetButton)
        qt.QAction.__init__(self, icon, "Clear", parent)


class _AddAction(qt.QAction):
    def __init__(self, parent):
        style = qt.QApplication.instance().style()
        icon = style.standardIcon(qt.QStyle.SP_DirIcon)
        qt.QAction.__init__(self, icon, "Add", parent)


class SliceStack(_ImageStack):
    """
    Widget displaying all slices contained in a list of acquisition folder
    """

    def extractImages(self):
        """
        Parse all self._scans and find the images to be displayed on the widget

        :return: images to display for each scan
        :rtype: dict
        """
        slices = {}
        for scan in self._scans:
            if not isinstance(scan, TomwerScanBase):
                scan_obj = ScanFactory.create_scan_object(scan_path=scan)
            else:
                scan_obj = scan
            if (
                scan_obj.latest_reconstructions is not None
                and len(scan_obj.latest_reconstructions) > 0
            ):

                def get_urls(volume_id: VolumeIdentifier):
                    volume = Factory.create_tomo_object_from_identifier(volume_id)
                    return list(volume.browse_data_urls())

                urls = list()
                for volume_id in scan_obj.latest_reconstructions:
                    urls.extend(get_urls(volume_id))

                slices[str(scan)] = dict(enumerate(urls))
            else:
                imgs = dict(enumerate(scan_obj.get_reconstructions_urls()))
                if len(imgs) > 0:
                    slices[str(scan)] = imgs
        return slices


class RadioStack(_ImageStack):
    """
    Widget displaying all radio contained in a list of acquisition folder
    """

    def extractImages(self):
        """
        Parse all self._scans and find the images to be displayed on the widget

        :return: images to display for each scan
        :rtype: dict
        """
        slices = {}
        for scan in self._scans:
            # manage hdf5
            if isinstance(scan, TomwerScanBase):
                imgs = scan.projections
                if len(imgs) > 0:
                    slices[str(scan)] = imgs
            elif h5py.is_hdf5(scan):
                try:
                    scans = ScanFactory.create_scan_objects(scan_path=scan)
                except Exception as e:
                    _logger.warning(e)
                else:
                    for scan_ in scans:
                        imgs = scan_.projections
                        if len(imgs) > 0:
                            slices[str(scan_)] = imgs
            else:
                # manage edf
                try:
                    scan_ = ScanFactory.create_scan_object(scan_path=scan)
                except ValueError:
                    pass
                else:
                    imgs = scan_.projections
                    if len(imgs) > 0:
                        slices[scan] = imgs
        return slices



"""
Button of general usage.
"""

import numpy
import logging
from typing import Optional, Union
from silx.gui import qt
from silx.gui.plot.tools.roi import RegionOfInterestManager
from silx.gui.plot.items.roi import LineROI
from silx.gui.plot.PlotToolButtons import PlotToolButton

from tomwer.gui import icons
from pyunitsystem.metricsystem import MetricSystem


_logger = logging.getLogger(__file__)


class PadlockButton(qt.QPushButton):
    """Simple button to define a button with PadLock icons"""

    sigLockChanged = qt.Signal(bool)
    """signal emitted when the lock status change"""

    def __init__(self, parent):
        qt.QPushButton.__init__(self, parent)
        self._lockIcon = icons.getQIcon("locked")
        self._unlockIcon = icons.getQIcon("unlocked")
        self.setIcon(self._unlockIcon)
        self.setCheckable(True)

        # connect signals
        self.toggled.connect(self._updateDisplay)

    def setLock(self, lock: bool):
        self.setChecked(lock)
        self._updateDisplay(lock)

    def _updateDisplay(self, checked):
        _icon = self._lockIcon if checked else self._unlockIcon
        self.setIcon(_icon)
        self.sigLockChanged.emit(checked)

    def isLocked(self) -> bool:
        return self.isChecked()


class TabBrowsersButtons(qt.QWidget):
    """Simple widget containing buttons to go to 'next' or 'previous'"""

    sigPreviousReleased = qt.Signal()
    """emitted when the previous button is released"""
    sigNextReleased = qt.Signal()
    """emitted when the next button is released"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        # define gui
        style = qt.QApplication.style()
        self._previousButton = qt.QPushButton("previous", self)
        previous_icon = style.standardIcon(qt.QStyle.SP_ArrowLeft)
        self._previousButton.setIcon(previous_icon)
        self.layout().addWidget(self._previousButton)

        self._nextButton = qt.QPushButton("next", self)
        next_icon = style.standardIcon(qt.QStyle.SP_ArrowRight)
        self._nextButton.setIcon(next_icon)
        self.layout().addWidget(self._nextButton)
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(spacer)

        # connect signal / slot
        self._previousButton.released.connect(self._previousReleased)
        self._nextButton.released.connect(self._nextReleased)

    def _previousReleased(self, *args, **kwargs):
        self.sigPreviousReleased.emit()

    def _nextReleased(self, *args, **kwargs):
        self.sigNextReleased.emit()


class TapeMeasureToolButton(PlotToolButton):
    """Button to active measurement between two point of the plot"""

    class TapeMeasureROI(LineROI):
        def __init__(self, parent=None, pixel_size_m=None):
            super().__init__(parent)
            self._pixel_size_m = None
            self.setPixelSize(pixel_size_m)

        def setEndPoints(self, startPoint, endPoint):
            distance_px = numpy.linalg.norm(endPoint - startPoint)
            super().setEndPoints(startPoint=startPoint, endPoint=endPoint)
            if self._pixel_size_m is None:
                self._updateText(f"{distance_px :.1f}px")
            else:
                distance_m = distance_px * self._pixel_size_m
                value, unit = MetricSystem.cast_metric_to_best_unit(distance_m)
                self._updateText(f"{distance_px :.1f}px ({value:.2f}{unit})")

        def setPixelSize(self, pixel_size_m: Optional[Union[tuple, float]]):
            if isinstance(pixel_size_m, (tuple, list)):
                assert (
                    len(pixel_size_m) == 2
                ), "expects at most two pixel size values (x and y values)"
                if not numpy.isclose(pixel_size_m[0], pixel_size_m[1]):
                    value, unit = MetricSystem.cast_metric_to_best_unit(pixel_size_m[0])
                    _logger.warning(
                        f"TapeMeasure is only handling square pixels for now. Will consider the pixel is {value:.2f}{unit}x{value:.2f}{unit}"
                    )
                pixel_size_m = pixel_size_m[0]
            self._pixel_size_m = pixel_size_m

    def __init__(self, parent=None, plot=None, pixel_size_mm=None):
        super().__init__(parent=parent, plot=plot)
        self._roiManager = None
        self._lastRoiCreated = None
        self._pixel_sixel_m = pixel_size_mm
        self.setIcon(icons.getQIcon("ruler"))
        self.setToolTip("measure distance between two pixels")
        self.toggled.connect(self._callback)
        self._connectPlot(plot)

    def setPlot(self, plot):
        return super().setPlot(plot)

    def setPixelSize(self, pixel_size_m: Optional[Union[tuple, float]]):
        self._pixel_sixel_m = pixel_size_m

    def _callback(self, toggled):
        if not self._roiManager:
            return
        if self._lastRoiCreated is not None:
            self._lastRoiCreated.setVisible(self.isChecked())
        if self.isChecked():
            self._roiManager.start(
                self.TapeMeasureROI,
                self,
            )
            self.__interactiveModeStarted(self._roiManager)
        else:
            source = self._roiManager.getInteractionSource()
            if source is self:
                self._roiManager.stop()

    def __interactiveModeStarted(self, roiManager):
        roiManager.sigInteractiveModeFinished.connect(self.__interactiveModeFinished)

    def __interactiveModeFinished(self):
        roiManager = self._roiManager
        if roiManager is not None:
            roiManager.sigInteractiveModeFinished.disconnect(
                self.__interactiveModeFinished
            )
        self.setChecked(False)

    def _connectPlot(self, plot):
        """
        Called when the plot is connected to the widget
        :param plot: :class:`.PlotWidget` instance
        """
        if plot is None:
            return
        self._roiManager = RegionOfInterestManager(plot)
        self._roiManager.setColor("yellow")  # Set the color of ROI
        self._roiManager.sigRoiAdded.connect(self._registerCurrentROI)

    def _disconnectPlot(self, plot):
        if plot and self._lastRoiCreated is not None:
            self._roiManager.removeRoi(self._lastRoiCreated)
            self._lastRoiCreated = None
        return super()._disconnectPlot(plot)

    def _registerCurrentROI(self, currentRoi):
        if self._lastRoiCreated is None:
            self._lastRoiCreated = currentRoi
            self._lastRoiCreated.setPixelSize(self._pixel_sixel_m)
        elif currentRoi != self._lastRoiCreated and self._roiManager is not None:
            self._roiManager.removeRoi(self._lastRoiCreated)
            self._lastRoiCreated = currentRoi
            self._lastRoiCreated.setPixelSize(self._pixel_sixel_m)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "12/12/2018"


import logging
import time

from tomwer.core.process.task import Task
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.core.utils.deprecation import deprecated_warning, deprecated

_logger = logging.getLogger(__name__)


class TimerTask(
    Task,
    input_names=("data", "wait"),
    output_names=("data",),
    optional_input_names=("serialize_output_data",),
):
    """
    Simple timer / time out - function
    """

    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        if inputs is None:
            inputs = {}

    @property
    @deprecated(replacement="task.wait", since_version="1.2")
    def waiting_time(self):
        return self.inputs.wait

    @waiting_time.setter
    @deprecated(replacement="task.wait", since_version="1.2")
    def waiting_time(self, wait):
        self.inputs.wait = wait

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if type(scan) is dict:
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            scan = scan
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be a dict or an instance of TomwerScanBase. Not {type(scan)}"
            )
        time.sleep(self.inputs.wait)
        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan


class Timer(TimerTask):
    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        deprecated_warning(
            name="tomwer.core.process.control.timer.Timer",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="TimerTask",
        )
        super().__init__(varinfo, inputs, node_id, node_attrs, execinfo)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "05/04/2019"


import os
import shutil
import tempfile
import unittest

import numpy
from tomoscan.esrf.scan.utils import get_data

from tomwer.core.process.reconstruction.darkref.params import DKRFRP
from tomwer.core.process.reconstruction.darkref.params import ReduceMethod as cMethod
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.scanutils import MockEDF, MockNXtomo
from tomwer.tests.datasets import TomwerCIDatasets

from ..reconstruction.darkref.darkrefs import DarkRefsTask
from ..reconstruction.darkref.darkrefscopy import DarkRefsCopy


class TestDarkRefIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    def setUp(self):
        self.scan_folder = tempfile.mkdtemp()

        self.scan_edf = MockEDF.mockScan(
            scanID=self.scan_folder, nRadio=10, nRecons=1, nPagRecons=4, dim=10
        )
        self.mock_hdf5 = MockNXtomo(
            scan_path=self.scan_folder, n_proj=10, n_pag_recons=0
        )
        self.scan_hdf5 = self.mock_hdf5.scan

        self.recons_params = DKRFRP()

    def tearDown(self):
        shutil.rmtree(self.scan_folder)

    def testInputOutput(self):
        for scan, scan_type in zip(
            (self.scan_edf, self.scan_hdf5), ("edf scan", "hdf5 scan")
        ):
            for input_type in (dict, TomwerScanBase):
                for serialize_output_data in (True, False):
                    with self.subTest(
                        return_dict=serialize_output_data,
                        input_type=input_type,
                        scan_type=scan_type,
                    ):
                        input_obj = scan
                        if input_obj is dict:
                            input_obj = input_obj.to_dict()
                        process = DarkRefsTask(
                            inputs={
                                "dark_ref_params": self.recons_params,
                                "data": input_obj,
                                "serialize_output_data": serialize_output_data,
                            }
                        )
                        process.run()
                        out = process.outputs.data
                        if serialize_output_data:
                            self.assertTrue(isinstance(out, dict))
                        else:
                            self.assertTrue(isinstance(out, TomwerScanBase))


class TestDarkRefCopyIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    def setUp(self):
        self.scan_folder = tempfile.mkdtemp()

        self.scan_edf = MockEDF.mockScan(
            scanID=self.scan_folder, nRadio=10, nRecons=1, nPagRecons=4, dim=10
        )
        self.scan_hdf5 = MockNXtomo(
            scan_path=self.scan_folder, n_proj=10, n_pag_recons=0
        ).scan
        self.recons_params = DKRFRP()
        self._save_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.scan_folder)

    def testInputOutput(self):
        for scan, scan_type in zip(
            (self.scan_edf, self.scan_hdf5), ("edf scan", "hdf5 scan")
        ):
            for input_type in (dict, TomwerScanBase):
                for serialize_output_data in (True, False):
                    with self.subTest(
                        return_dict=serialize_output_data,
                        input_type=input_type,
                        scan_type=scan_type,
                    ):
                        input_obj = scan
                        if input_obj is dict:
                            input_obj = input_obj.to_dict()
                        dkrf_process = DarkRefsCopy(
                            inputs={
                                "dark_ref_params": self.recons_params,
                                "data": input_obj,
                                "serialize_output_data": serialize_output_data,
                                "save_dir": self._save_dir,
                            }
                        )
                        dkrf_process.run()
                        out = dkrf_process.outputs.data
                        if serialize_output_data:
                            self.assertTrue(isinstance(out, dict))
                        else:
                            self.assertTrue(isinstance(out, TomwerScanBase))


class TestDarkRefEdf(unittest.TestCase):
    """
    Test dark ref is correctly processing with h5 refs
    """

    def setUp(self) -> None:
        self.scan_folder = tempfile.mkdtemp()
        self.dark_data = numpy.array(  # pylint: disable=E1121
            numpy.random.random(4 * 200 * 200) * 100,
            dtype=numpy.uint16,
        ).reshape(4, 200, 200)
        self.flat_data = numpy.array(  # pylint: disable=E1121
            numpy.random.random(4 * 200 * 200) * 100,
            dtype=numpy.uint32,
        ).reshape(4, 200, 200)
        self.scan = MockEDF.mockScan(
            scanID=self.scan_folder,
            nRadio=10,
            nRecons=1,
            nPagRecons=4,
            dim=200,
            start_dark=True,
            start_flat=True,
            start_dark_data=self.dark_data,
            start_flat_data=self.flat_data,
        )
        self.recons_params = DKRFRP()
        self.recons_params.overwrite_dark = True
        self.recons_params.overwrite_ref = True
        self.assertFalse(os.path.exists(self.scan.process_file))

    def tearDown(self) -> None:
        shutil.rmtree(self.scan_folder)

    def testDark(self):
        """
        Test darks are computed when only dark are requested
        """
        method_to_test = (
            cMethod.MEDIAN,
            cMethod.MEAN,
            cMethod.FIRST,
            cMethod.LAST,
        )
        th_results = (
            numpy.median(self.dark_data, axis=0),
            numpy.mean(self.dark_data, axis=0),
            self.dark_data[0],
            self.dark_data[-1],
        )
        for method, th_res in zip(method_to_test, th_results):
            with self.subTest(method=method):
                if os.path.exists(self.scan.process_file):
                    os.remove(self.scan.process_file)
                process = DarkRefsTask(
                    inputs={
                        "dark_ref_params": self.recons_params,
                        "data": self.scan,
                        "serialize_output_data": False,
                    }
                )

                self.recons_params.dark_calc_method = method
                self.recons_params.flat_calc_method = cMethod.NONE
                process.run()
                numpy.testing.assert_array_almost_equal(
                    self.scan.reduced_darks[0], th_res.astype(numpy.uint16)
                )

    def testFlat(self):
        """
        Test flats are computed when only flat are requested
        """
        method_to_test = (
            cMethod.MEDIAN,
            cMethod.MEAN,
            cMethod.FIRST,
            cMethod.LAST,
        )
        th_results = (
            numpy.median(self.flat_data, axis=0),
            numpy.mean(self.flat_data, axis=0),
            self.flat_data[0],
            self.flat_data[-1],
        )
        for method, th_res in zip(method_to_test, th_results):
            with self.subTest(method=method):
                if os.path.exists(self.scan.process_file):
                    os.remove(self.scan.process_file)
                self.recons_params.dark_calc_method = cMethod.NONE
                self.recons_params.flat_calc_method = method
                process = DarkRefsTask(
                    inputs={
                        "dark_ref_params": self.recons_params,
                        "data": self.scan,
                        "serialize_output_data": False,
                    }
                )
                process.run()
                numpy.testing.assert_array_almost_equal(
                    self.scan.reduced_flats[0], th_res.astype(numpy.uint16)
                )


class TestDarkRefNx(unittest.TestCase):
    """
    Test dark ref is correctly processing with h5 refs
    """

    def setUp(self) -> None:
        super().setUp()
        dataset_name = "frm_edftomomill_twoentries.nx"
        self.scan_folder = tempfile.mkdtemp()
        self._file_path = os.path.join(self.scan_folder, dataset_name)
        shutil.copyfile(
            src=TomwerCIDatasets.get_dataset(f"h5_datasets/{dataset_name}"),
            dst=self._file_path,
        )
        self.scan = NXtomoScan(scan=self._file_path, entry="entry0000")
        self.assertFalse(os.path.exists(self.scan.process_file))
        self.recons_params = DKRFRP()
        self.recons_params.overwrite_dark = True
        self.recons_params.overwrite_ref = True

    def tearDown(self) -> None:
        shutil.rmtree(self.scan_folder)
        super().tearDown()

    def testDark(self):
        """
        Test darks are computed when only dark are requested
        """
        darks = self.scan.darks
        self.assertEqual(len(darks), 1)

        method_to_test = (cMethod.MEAN, cMethod.MEDIAN, cMethod.FIRST, cMethod.LAST)
        for method in method_to_test:
            with self.subTest(method=method):
                self.recons_params.dark_calc_method = method
                self.recons_params.flat_calc_method = cMethod.NONE
                process = DarkRefsTask(
                    inputs={
                        "dark_ref_params": self.recons_params,
                        "data": self.scan,
                        "serialize_output_data": False,
                    }
                )
                process.run()
                flats = self.scan.load_reduced_flats()
                self.assertEqual(len(flats), 0)
                darks = self.scan.load_reduced_darks()
                self.assertEqual(len(darks), 1)

    def testFlat(self):
        """
        Test flats are computed when only flat are requested
        """
        flats = self.scan.flats
        self.assertEqual(len(flats), 42)
        url_flat_serie_1 = [flats[index] for index in range(1, 22)]
        url_flat_serie_2 = [flats[index] for index in range(1521, 1542)]
        assert len(url_flat_serie_1) == 21
        assert len(url_flat_serie_2) == 21
        data_flat_serie_1 = [get_data(url) for url in url_flat_serie_1]
        data_flat_serie_2 = [get_data(url) for url in url_flat_serie_2]

        self.recons_params.overwrite_dark = True
        self.recons_params.overwrite_ref = True

        method_to_test = (cMethod.MEAN, cMethod.MEDIAN, cMethod.FIRST, cMethod.LAST)
        for method in method_to_test:
            with self.subTest(method=method):
                if method is cMethod.MEDIAN:
                    expected_res_s1 = numpy.median(data_flat_serie_1, axis=0)
                    expected_res_s2 = numpy.median(data_flat_serie_2, axis=0)
                elif method is cMethod.MEAN:
                    expected_res_s1 = numpy.mean(data_flat_serie_1, axis=0)
                    expected_res_s2 = numpy.mean(data_flat_serie_2, axis=0)
                elif method is cMethod.FIRST:
                    expected_res_s1 = data_flat_serie_1[0]
                    expected_res_s2 = data_flat_serie_2[0]
                elif method is cMethod.LAST:
                    expected_res_s1 = data_flat_serie_1[-1]
                    expected_res_s2 = data_flat_serie_2[-1]
                else:
                    raise ValueError("method not managed")

                self.recons_params.dark_calc_method = cMethod.NONE
                self.recons_params.flat_calc_method = method
                process = DarkRefsTask(
                    inputs={
                        "dark_ref_params": self.recons_params,
                        "data": self.scan,
                        "serialize_output_data": False,
                    }
                )
                process.run()

                darks = self.scan.load_reduced_darks()
                self.assertEqual(len(darks), 0)
                flats = self.scan.load_reduced_flats()
                self.assertEqual(len(flats), 2)
                self.assertTrue(1 in flats)
                self.assertTrue(1521 in flats)
                self.assertTrue(numpy.allclose(flats[1], expected_res_s1))
                self.assertTrue(numpy.allclose(flats[1521], expected_res_s2))

    def testDarkAndFlat(self):
        """
        Test darks and flats are computed when both are requested
        """
        flats = self.scan.flats
        self.assertEqual(len(flats), 42)
        url_flat_serie_1 = [flats[index] for index in range(1, 22)]
        url_flat_serie_2 = [flats[index] for index in range(1521, 1542)]
        assert len(url_flat_serie_1) == 21
        assert len(url_flat_serie_2) == 21
        data_flat_serie_1 = [get_data(url) for url in url_flat_serie_1]
        data_flat_serie_2 = [get_data(url) for url in url_flat_serie_2]

        self.recons_params.dark_calc_method = cMethod.MEAN
        self.recons_params.flat_calc_method = cMethod.MEDIAN
        darks = self.scan.darks
        self.assertEqual(len(darks), 1)
        dark_data = get_data(list(darks.values())[0])

        expected_flats_s1 = numpy.median(data_flat_serie_1, axis=0)
        expected_flats_s2 = numpy.median(data_flat_serie_2, axis=0)
        process = DarkRefsTask(
            inputs={
                "dark_ref_params": self.recons_params,
                "data": self.scan,
                "serialize_output_data": False,
            }
        )
        process.run()
        darks = self.scan.load_reduced_darks()
        flats = self.scan.load_reduced_flats()
        self.assertEqual(len(darks), 1)
        self.assertEqual(len(flats), 2)
        self.assertTrue(0 in darks)
        self.assertTrue(1 in flats)
        self.assertTrue(1521 in flats)
        self.assertTrue(numpy.allclose(flats[1], expected_flats_s1))
        self.assertTrue(numpy.allclose(flats[1521], expected_flats_s2))
        self.assertTrue(numpy.allclose(darks[0], dark_data))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/11/2020"


import nxtomomill.version
from nxtomomill.utils import add_dark_flat_nx_file
from silx.io.url import DataUrl
from tomwer.core.utils.deprecation import deprecated_warning, deprecated

from tomwer.core.process.task import Task
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.scanutils import data_identifier_to_scan


def apply_dark_flat_patch(scan: NXtomoScan, config: dict) -> TomwerScanBase:
    """

    :param scan:
    :param config:
    :return:
    """
    if not isinstance(scan, NXtomoScan):
        raise ValueError(
            f"Dark and flat patch only manage NXtomoScan and not {type(scan)}"
        )
    if config is None:
        return scan
    for param in ("darks_start", "darks_end", "flats_start", "flats_end"):
        if param not in config:
            config[param] = None

    add_dark_flat_nx_file(
        file_path=scan.master_file,
        entry=scan.entry,
        **config,
    )
    return scan


class DarkFlatPatchTask(
    Task,
    input_names=("data", "configuration"),
    output_names=("data",),
    optional_input_names=("serialize_output_data",),
):
    """
    Patch an existing NXtomo calling nxtomomill
    """

    def run(self):
        scan = data_identifier_to_scan(self.inputs.data)
        if scan is None:
            return
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be a dict or an instance of TomwerScanBase. Not {type(scan)}"
            )
        if not isinstance(scan, NXtomoScan):
            raise ValueError(f"input type of {scan}: {type(scan)} is not managed")

        config = self.inputs.configuration
        if not isinstance(config, dict):
            raise TypeError(f"config is expected to be a dict. {type(config)} provided")
        apply_dark_flat_patch(scan=scan, config=config)
        keys = config.keys()
        for key in keys:
            value = config[key]
            if isinstance(value, DataUrl):
                config[key] = value.path()

        self.register_process(
            process_file=scan.process_file,
            entry=scan.entry,
            configuration=config,
            results={},
            process_index=scan.pop_process_index(),
            overwrite=True,
        )
        if self.get_input_value("serialize_output_data", True):
            self.outputs.data = scan.to_dict()
        else:
            self.outputs.data = scan

    @staticmethod
    def program_name():
        return "nxtomomill.utils.change_image_key_control"

    @staticmethod
    def program_version():
        return nxtomomill.version.version

    @staticmethod
    def definition():
        return "Apply patch for dark and references on a scan (TomwerScanBase)"

    @deprecated(
        since_version="1.2",
        replacement="DarkFlatPatchTask.inputs.configuration",
        reason="ewoksification",
    )
    def get_configuration(self):
        """

        :return: configuration of the process
        :rtype: dict
        """
        return self.inputs.configuration

    @deprecated(
        since_version="1.2",
        replacement="DarkFlatPatchTask.inputs.configuration",
        reason="ewoksification",
    )
    def set_configuration(self, configuration: dict) -> None:
        self.inputs.configuration = configuration


class DarkFlatPatch(DarkFlatPatchTask):
    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        deprecated_warning(
            name="tomwer.core.process.edit.darkflatpatch.DarkFlatPatch",
            type_="class",
            reason="improve readibility",
            since_version="1.2",
            replacement="DarkFlatPatchTask",
        )
        super().__init__(varinfo, inputs, node_id, node_attrs, execinfo)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "09/08/2018"


import os

import h5py

try:
    # new settings names
    from nxtomomill.settings import Tomo

    H5_INIT_TITLES = Tomo.H5.INIT_TITLES
    from nxtomomill.settings import Tomo

    H5_ZSERIE_INIT_TITLES = Tomo.H5.ZSERIE_INIT_TITLES

    H5_PCOTOMO_INIT_TITLES = Tomo.H5.PCOTOMO_INIT_TITLES

except ImportError:
    # old settings names
    # TODO: remove in 1.1
    from nxtomomill.settings import H5_INIT_TITLES
    from nxtomomill.settings import H5_ZSERIE_INIT_TITLES
    from nxtomomill.settings import H5_PCOTOMO_INIT_TITLES

import logging
from typing import Optional

from silx.io.utils import h5py_read_dataset
from silx.io.utils import open as open_hdf5

_logger = logging.getLogger(__name__)


class BlissScan:
    """Simple class to define a Bliss sequence aka as Bliss scan inside tomwer.

    :warning: BlissScan is not compatible with tomwer treatment. This is
    why it does not inherit from TomwerScanBase. This is a utility class.
    """

    _TYPE = "bliss_hdf5"

    def __init__(
        self, master_file, entry, proposal_file, scan_numbers=None, saving_file=None
    ):
        self._master_file = master_file
        if isinstance(entry, str) and not entry.startswith("/"):
            self._entry = "/" + entry
        else:
            self._entry = entry
        self._proposal_file = proposal_file
        self._scan_numbers = scan_numbers or []
        self._saving_file = saving_file
        self._tomo_n = None
        self._n_acquired = None
        self._dir_path = os.path.dirname(self.master_file)

    @property
    def tomo_n(self):
        """total number of projections"""
        return self._tomo_n

    @property
    def proposal_file(self):
        return self._proposal_file

    @property
    def saving_file(self):
        return self._saving_file

    @tomo_n.setter
    def tomo_n(self, n):
        self._tomo_n = n

    @property
    def n_acquired(self):
        """
        number of frame acquired until now. Does not take into account
        dark, flats or alignment"""
        return self._n_acquired

    @n_acquired.setter
    def n_acquired(self, n):
        self._n_acquired = n

    @property
    def master_file(self):
        return self._master_file

    @property
    def entry(self):
        return self._entry

    @property
    def path(self):
        return self._dir_path

    @property
    def scan_numbers(self):
        return self._scan_numbers

    def add_scan_number(self, scan_number):
        self._scan_numbers.append(scan_number)

    def __str__(self):
        return self.get_id_name(master_file=self.master_file, entry=self.entry)

    @staticmethod
    def get_id_name(master_file, entry):
        return "@".join((str(entry), master_file))

    def _deduce_transfert_scan(self, output_dir):
        new_master_file = os.path.join(
            output_dir,
            os.path.basename(os.path.dirname(self.master_file)),
            os.path.basename(self.master_file),
        )
        new_proposal_file = os.path.join(
            output_dir, os.path.basename(self.proposal_file)
        )
        return BlissScan(
            master_file=new_master_file,
            proposal_file=new_proposal_file,
            entry=self.entry,
        )

    @staticmethod
    def is_bliss_file(file_path):
        return len(BlissScan.get_valid_entries(file_path)) > 0

    def is_abort(self, *args, **kwargs):
        # for now there is not real way to know if a scan is abort or not
        return False

    def scan_dir_name(self) -> Optional[str]:
        """for 'this/is/my/acquisition' returns 'acquisition'"""
        if self.path is not None:
            return self.path.split(os.sep)[-1]
        else:
            return None

    def scan_basename(self) -> Optional[str]:
        """for 'this/is/my/acquisition' returns 'acquisition'"""
        if self.path is not None:
            return self.path
        else:
            return None

    def scan_parent_dir_basename(self) -> Optional[str]:
        if self.path is not None:
            try:
                return os.path.dirname(self.path)
            except Exception:
                return None
        else:
            return None

    @staticmethod
    def is_bliss_valid_entry(file_path: str, entry: str):
        def is_pcotmo_tomo(current_title: str):
            for title in H5_PCOTOMO_INIT_TITLES:
                if current_title.startswith(title):
                    return True
            return False

        def is_standard_tomo(current_title: str):
            for title in H5_INIT_TITLES:
                if current_title.startswith(title):
                    return True
            return False

        def is_zseries_tomo(current_title: str):
            for title in H5_ZSERIE_INIT_TITLES:
                if current_title.startswith(title):
                    return True
            return False

        with open_hdf5(file_path) as h5s:
            if not isinstance(h5s, h5py.Group) or not isinstance(
                h5s[entry], h5py.Group
            ):
                return False
            if "title" in h5s[entry]:
                title = h5py_read_dataset(h5s[entry]["title"])
                try:
                    return (
                        is_standard_tomo(title)
                        or is_zseries_tomo(title)
                        or is_pcotmo_tomo(title)
                    )
                except Exception:
                    pass
            return False

    @staticmethod
    def get_valid_entries(file_path) -> tuple:
        if not h5py.is_hdf5(file_path):
            _logger.warning("Provided file %s is not a hdf5 file" % file_path)
            return tuple()
        else:
            res = []
            with open_hdf5(file_path) as h5s:
                for entry in h5s:
                    if BlissScan.is_bliss_valid_entry(file_path=file_path, entry=entry):
                        res.append(entry)
            return tuple(res)

    def to_dict(self):
        return {
            "DICT_TYPE_KEY": self._TYPE,
            "master_file": self.master_file,
            "entry": self.entry,
            "proposal_file": self.proposal_file,
            "scan_numbers": self.scan_numbers,
        }

    @staticmethod
    def from_dict(ddict):
        master_file = ddict["master_file"]
        entry = ddict["entry"]
        return BlissScan(
            master_file=master_file, entry=entry, proposal_file=None
        ).load_frm_dict(ddict=ddict)

    def load_frm_dict(self, ddict):
        if "master_file" in ddict:
            self._master_file = ddict["master_file"]
        if "entry" in ddict:
            self._entry = ddict["entry"]
        if "proposal_file" in ddict:
            self._proposal_file = ddict["proposal_file"]
        if "scan_numbers" in ddict:
            self._scan_numbers = ddict["scan_numbers"]
        return self



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "12/01/2022"


import logging
from contextlib import AbstractContextManager
from typing import Iterable, Optional

from silx.gui import qt
from silx.gui.utils import blockSignals
from silx.utils.enum import Enum as _Enum
from tomoscan.serie import Serie

from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.gui.control.datalist import TomoObjList
from tomwer.gui.qfolderdialog import QDataDialog
from tomwer.gui.visualization.tomoobjoverview import TomoObjOverview

_logger = logging.getLogger(__name__)


class SerieWidgetDialog(qt.QDialog):
    sigSerieSelected = qt.Signal(Serie)
    """
    emit when a serie is selected / triggered by the user
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.setLayout(qt.QVBoxLayout())
        # add list
        self._widget = SerieWidget()
        self.layout().addWidget(self._widget)
        # add buttons
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._selectButton = qt.QPushButton("Select (active) serie", parent=self)
        self._buttons.addButton(self._selectButton, qt.QDialogButtonBox.ActionRole)
        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self._selectButton.released.connect(self._serieSelected)

    def getSelectedSerie(self) -> Optional[Serie]:
        return self._widget.getSelectedSerie()

    def _serieSelected(self, *args, **kwargs):
        serie = self.getSelectedSerie()
        if serie is not None:
            self.sigSerieSelected.emit(serie)

    # expose API
    def add(self, tomo_obj):
        self._widget.add(tomo_obj=tomo_obj)


class SerieWidget(qt.QTabWidget):
    sigCurrentSerieChanged = qt.Signal()
    """signal emit when the current serie changes"""

    sigHistoryChanged = qt.Signal()
    """signal emit when the history changed (a serie has been added or removed"""

    sigSerieSend = qt.Signal(Serie)
    """Signal emited when a serie has been send"""

    _HISTORY_MODE = "history"
    _DEFINITION_MODE = "serie definition"

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setWindowTitle("serie of scans")
        self._serieDefinitionWidget = SerieDefinition(parent=self)
        self.addTab(self._serieDefinitionWidget, self._DEFINITION_MODE)
        self._historyWidget = SerieHistoryDialog(parent=self)
        self._historyWidget.setWindowFlags(qt.Qt.Widget)
        self.addTab(self._historyWidget, self._HISTORY_MODE)

        # conenct signal / slot
        self._historyWidget.sigEditSerie.connect(self._serieEditionRequested)
        self._historyWidget.sigSerieSend.connect(self.sigSerieSend)
        self._historyWidget.sigHistoryUpdated.connect(self._repeatHistoryUpdated)
        self._serieDefinitionWidget.sigSerieChanged.connect(self._repeatSerieChanged)
        self._serieDefinitionWidget.sigSerieSend.connect(self.sigSerieSend)
        self._serieDefinitionWidget.sigSerieSend.connect(self._historyWidget.addSerie)

    def getHistoryWidget(self):
        return self._historyWidget

    def getDefinitionWidget(self):
        return self._serieDefinitionWidget

    def getSelectedSerie(self) -> Optional[Serie]:
        return self._serieDefinitionWidget.getSelectedSerie()

    def setMode(self, mode: str, definition_mode: Optional[str] = None):
        valid_modes = (self._HISTORY_MODE, self._DEFINITION_MODE)
        if mode == self._HISTORY_MODE:
            self.setCurrentWidget(self._historyWidget)
        elif mode == self._DEFINITION_MODE:
            self.setCurrentWidget(self._serieDefinitionWidget)
            self._serieDefinitionWidget.setMode(definition_mode)
        else:
            raise ValueError(
                f"mode {mode} is no recognized. Valid modes are {valid_modes}"
            )

    def _serieEditionRequested(self, serie: Serie):
        if not isinstance(serie, Serie):
            raise TypeError(f"serie is expected to be a serie not {type(serie)}")
        self.setMode("serie definition", "manual")
        self.getDefinitionWidget().getManualDefinitionWidget().setSerie(serie)

    def _repeatSerieChanged(self, *args, **kwargs):
        self.sigCurrentSerieChanged.emit()

    def _repeatHistoryUpdated(self, *args, **kwargs):
        self.sigHistoryChanged.emit()

    def add(self, tomo_obj):
        return self._serieDefinitionWidget.addTomoObj(tomo_obj)


class _SerieDefinitionMode(_Enum):
    MANUAL = "manual"
    AUTO = "auto"


class SerieDefinition(qt.QWidget):
    sigSerieChanged = qt.Signal()
    ### signal emit when a the serie defined manually changed

    sigSerieSend = qt.Signal(Serie)
    ### signal emit when a serie is send

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QGridLayout())

        self._modeLabel = qt.QLabel("Mode", self)
        self.layout().addWidget(self._modeLabel, 0, 0, 1, 1)

        self._modeCB = qt.QComboBox(self)
        for mode in _SerieDefinitionMode.values():
            self._modeCB.addItem(mode)
        self.layout().addWidget(self._modeCB, 0, 1, 1, 1)

        self._manualDefWidget = SerieManualFromTomoObj(parent=self)
        self.layout().addWidget(self._manualDefWidget, 1, 0, 1, 2)
        self._manualDefWidget.setWindowFlags(qt.Qt.Widget)

        self._automaticDefWidget = SerieAutomaticDefinitionWidget(parent=self)
        self.layout().addWidget(self._automaticDefWidget, 2, 0, 1, 2)

        # connect signal / slot
        self._modeCB.currentIndexChanged.connect(self._updateVisibility)
        self._manualDefWidget._newSerieWidget.sigUpdated.connect(self.sigSerieChanged)

        # set up
        self._updateVisibility()

    def getSelectedSerie(self) -> Optional[Serie]:
        if self.getMode() == _SerieDefinitionMode.MANUAL:
            return self._manualDefWidget.getSerie()
        else:
            raise ValueError(f"mode {self.getMode()} is not handled yet")

    def getMode(self) -> str:
        return _SerieDefinitionMode.from_value(self._modeCB.currentText())

    def setMode(self, mode: str):
        mode = _SerieDefinitionMode.from_value(mode)
        idx = self._modeCB.findText(mode.value)
        self._modeCB.setCurrentIndex(idx)

    def _updateVisibility(self):
        self._manualDefWidget.setVisible(self.getMode() == _SerieDefinitionMode.MANUAL)
        self._automaticDefWidget.setVisible(self.getMode() == _SerieDefinitionMode.AUTO)

    def getManualDefinitionWidget(self):
        return self._manualDefWidget

    def getAutoDefinitionWidget(self):
        return self._automaticDefWidget

    def createManualSerie(self):
        self._manualDefWidget.createSerie()

    def addTomoObj(self, tomo_obj: TomwerObject):
        self._manualDefWidget.addTomoObj(tomo_obj=tomo_obj)

    def setSerieName(self, name: str):
        self._manualDefWidget.setSerieName(name=name)


class _SerieDefinitionTree(qt.QWidget):
    """
    Tree used to define manually serie of scan.
    Drag and drop of files is handled
    """

    sigUpdated = qt.Signal()
    """Signal emit when the serie is updated"""

    class SignalBlocker(AbstractContextManager):
        """Simple context manager to hide / show button dialogs"""

        def __init__(self, serie_definition_widget) -> None:
            super().__init__()
            self.serie_definition_widget = serie_definition_widget

        def __enter__(self):
            self.old_widget = self.serie_definition_widget.blockSignals(True)
            self.old_tree = self.serie_definition_widget._tree.blockSignals(True)

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.serie_definition_widget.blockSignals(self.old_widget)
            self.serie_definition_widget._tree.blockSignals(self.old_tree)

    def __init__(self, parent=None, serie_name="my_serie") -> None:
        self._tomo_objs = {}
        # associated serie name (key) to tuple (serie, QTreeWidgetItem)
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self._tree = qt.QTreeWidget(self)
        self._tree.setSelectionMode(qt.QAbstractItemView.MultiSelection)
        self._tree.setColumnCount(2)
        self._tree.setHeaderLabels(("serie", "scan ids"))
        self._tree.setItemsExpandable(False)
        self.layout().addWidget(self._tree)

        # set up the tree with the serie name that will stay during the entire
        # life time of the tree
        self._serieItem = qt.QTreeWidgetItem(self._tree)
        self._serieItem.setFlags(self._serieItem.flags() | qt.Qt.ItemIsEditable)
        self._serieItem.setExpanded(True)

        self.setAcceptDrops(True)

        # connect signal / slot
        self._tree.itemChanged.connect(self._updated)

        # set up
        self.setSerieName(name=serie_name)

        # expose API
        self.itemChanged = self._tree.itemChanged

    @property
    def rootItem(self):
        return self._serieItem

    def setSerieName(self, name: str):
        with self.SignalBlocker(self):
            self._serieItem.setText(0, name)
        self.sigUpdated.emit()

    def getSerieName(self):
        return self._serieItem.text(0)

    def addTomoObj(self, tomo_obj: TomwerObject):
        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"{tomo_obj} is expected to be an instance of {TomwerObject} not {type(tomo_obj)}"
            )
        identifier = tomo_obj.get_identifier().to_str()
        if identifier in self._tomo_objs:
            _logger.warning(f"scan {identifier} already part of the serie")
            return

        with self.SignalBlocker(self):
            tomo_obj_item = qt.QTreeWidgetItem(self.rootItem)
            tomo_obj_item.setText(1, identifier)
            tomo_obj_item.setFlags(tomo_obj_item.flags() | qt.Qt.ItemIsUserCheckable)
            self._tomo_objs[identifier] = (tomo_obj, tomo_obj_item)
        self.sigUpdated.emit()

    def removeTomoObj(self, tomo_obj: TomwerObject):
        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"{tomo_obj} is expected to be an instance of {TomwerObject} not {type(tomo_obj)}"
            )

        with self.SignalBlocker(self):
            identifier = tomo_obj.get_identifier().to_str()
            if identifier not in self._tomo_objs:
                _logger.warning(f"{identifier} is not in the serie")
            else:
                _, tomo_obj_item = self._tomo_objs.pop(identifier)
                root = self._tree.invisibleRootItem()
                root.removeChild(tomo_obj_item)
        self.sigUpdated.emit()

    @property
    def n_tomo_objs(self):
        return len(self._tomo_objs)

    def setSerie(self, serie: Serie) -> None:
        if not isinstance(serie, Serie):
            raise TypeError(
                f"serie is expected to be an instance of {Serie} not {type(serie)}"
            )

        with self.SignalBlocker(self):
            self.clearTomoObjs()
            self.setSerieName(serie.name)
            for tomo_obj in serie:
                if isinstance(tomo_obj, str):
                    try:
                        tomo_obj = ScanFactory.create_tomo_object_from_identifier(
                            identifier=tomo_obj
                        )
                    except Exception:
                        try:
                            tomo_obj = VolumeFactory.create_tomo_object_from_identifier(
                                identifier=tomo_obj
                            )
                        except Exception:
                            _logger.warning(f"Fail to recreate scan from {tomo_obj}.")
                            return
                elif not isinstance(tomo_obj, TomwerObject):
                    raise TypeError(
                        f"tomo_obj is expected to be an instance of {TomwerObject}. Not {type(tomo_obj)}"
                    )
                self.addTomoObj(tomo_obj)
        self.sigUpdated.emit()

    def getSerie(self, use_identifiers=False) -> Serie:
        scans = [scan for scan, _ in self._tomo_objs.values()]
        return Serie(
            name=self.getSerieName(),
            iterable=scans,
            use_identifiers=use_identifiers,
        )

    def clearTomoObjs(self):
        with self.SignalBlocker(self):
            keys = list(self._tomo_objs.keys())
            for key in keys:
                _, scan_item = self._tomo_objs.pop(key)
                root = self._tree.invisibleRootItem()
                root.removeChild(scan_item)
        self.sigUpdated.emit()

    def setSelectedTomoObjs(self, objs):
        self.clearSelection()
        for scan in objs:
            scan_item = self._getTomoObjItem(scan)
            if scan_item is not None:
                scan_item.setSelected(True)

    def _getTomoObjItem(self, tomo_obj: TomwerObject) -> Optional[qt.QTreeWidgetItem]:
        if not isinstance(tomo_obj, TomwerObject):
            raise TypeError(
                f"scan is expected to be an instance of {TomwerObject} not {type(tomo_obj)}"
            )
        return self._tomo_objs.get(tomo_obj.get_identifier().to_str(), (None, None))[1]

    def getSelectedTomoObjs(self) -> tuple():
        """return selected scans"""
        selected = []
        for _, (scan, item) in self._tomo_objs.items():
            if item.isSelected():
                selected.append(scan)
        return tuple(selected)

    def removeSelectedTomoObjs(self) -> None:
        with self.SignalBlocker(self):
            for tomo_obj in self.getSelectedTomoObjs():
                self.removeTomoObj(tomo_obj)

    def _updated(self, *args, **kwargs):
        self.sigUpdated.emit()

    def clearSelection(self) -> None:
        self._tree.selectionModel().clearSelection()

    def addScanFromNxFile(self, file_: str, entry: Optional[str] = None):
        try:
            if entry is None:
                scans = ScanFactory.create_scan_objects(scan_path=file_)
            else:
                scans = [ScanFactory.create_scan_object(scan_path=file_, entry=entry)]
        except Exception as e:
            _logger.error(f"cannot create scan instances from {file_}. Error is {e}")
        else:
            changed = False
            with self.SignalBlocker(self):
                for scan in scans:
                    if scan is not None:
                        try:
                            self.addTomoObj(tomo_obj=scan)
                        except TypeError:
                            _logger.error(
                                f"fail to add scan {scan}. Invalid type encountered ({type(scan)})"
                            )
                        else:
                            changed = True
            if changed:
                self.sigUpdated.emit()

    def dropEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            for url in event.mimeData().urls():
                self.addScanFromNxFile(file_=str(url.path()), entry=None)

    def supportedDropActions(self):
        """Inherited method to redefine supported drop actions."""
        return qt.Qt.CopyAction | qt.Qt.MoveAction

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            event.accept()
            event.setDropAction(qt.Qt.CopyAction)
        else:
            qt.QListWidget.dragEnterEvent(self, event)

    def dragMoveEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            event.setDropAction(qt.Qt.CopyAction)
            event.accept()
        else:
            qt.QListWidget.dragMoveEvent(self, event)


class SerieManualControlDialog(qt.QDialog):
    """
    Same as the :class:`SerieManualDefinitionDialog` but with control of the serie.
    This include a `create serie` and a `create serie and clear button`
    """

    sigSerieSend = qt.Signal(Serie)

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self._mainWidget = SerieManualDefinitionDialog(parent=self)
        self._mainWidget.setWindowFlags(qt.Qt.Widget)
        self.layout().addWidget(self._mainWidget)

        self._buttons = qt.QDialogButtonBox(parent=self)
        self._createButton = qt.QPushButton("create serie", parent=self)
        self._buttons.addButton(self._createButton, qt.QDialogButtonBox.ActionRole)

        # connect signal / slot
        self._createButton.clicked.connect(self._sendSerie)
        self.layout().addWidget(self._buttons)

        # expose API
        self.sigUpdated = self._mainWidget.sigUpdated

    def _sendSerie(self):
        self.sigSerieSend.emit(self._mainWidget.getSerie())

    @property
    def n_tomo_objs(self):
        return self._mainWidget.n_tomo_objs

    def setSerieName(self, name: str):
        self._mainWidget.setSerieName(name=name)

    def getSerieName(self) -> str:
        return self._mainWidget.getSerieName()

    def setSerie(self, serie: Serie) -> None:
        self._mainWidget.setSerie(serie)

    def getSerie(self, *args, **kwargs) -> Serie:
        return self._mainWidget.getSerie(*args, **kwargs)

    def addScanFromNxFile(self, file_: str, entry: Optional[str] = None):
        return self._mainWidget.addScanFromNxFile(file_=file_, entry=entry)

    def removeSelectedScans(self) -> None:
        return self._mainWidget.removeSelectedTomoObjs()

    def getSelectedScans(self) -> tuple:
        return self._mainWidget.getSelectedTomoObjs()

    def setSelectedScans(self, scans: Iterable) -> None:
        self._mainWidget.setSelectedTomoObjs(scans=scans)

    def addScan(self, scan: TomwerScanBase) -> None:
        self._mainWidget.addTomoObj(scan=scan)

    def removeScan(self, scan: TomwerScanBase) -> None:
        self._mainWidget.removeTomoObj(scan=scan)

    def clearSerie(self) -> None:
        self._mainWidget.clearSerie()

    def createSerie(self):
        self.sigSerieSend.emit(self.getSerie())


class SerieManualFromTomoObj(qt.QWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        style = qt.QApplication.style()
        self.setLayout(qt.QGridLayout())

        # tomo objs list
        self._tomoObjList = TomoObjList(self)
        self._tomoObjList.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding
        )
        self._tomoObjListScrollArea = qt.QScrollArea(self)
        self._tomoObjListScrollArea.setWidgetResizable(True)
        self._tomoObjListScrollArea.setWidget(self._tomoObjList)
        self.layout().addWidget(self._tomoObjListScrollArea, 0, 0, 4, 2)

        # right arrow
        self._rightArrowButton = qt.QPushButton(self)
        rightArrowIcon = style.standardIcon(qt.QStyle.SP_ArrowRight)
        self._rightArrowButton.setIcon(rightArrowIcon)
        self._rightArrowButton.setFixedWidth(30)
        self.layout().addWidget(self._rightArrowButton, 1, 2, 1, 1)

        # left arrow
        self._leftArrowButton = qt.QPushButton(self)
        leftArrowIcon = style.standardIcon(qt.QStyle.SP_ArrowLeft)
        self._leftArrowButton.setIcon(leftArrowIcon)
        self._leftArrowButton.setFixedWidth(30)
        self.layout().addWidget(self._leftArrowButton, 2, 2, 1, 1)

        # new serie
        self._newSerieWidget = NewSerieWidget(self)
        self._newSerieWidgetScrollArea = qt.QScrollArea(self)
        self._newSerieWidgetScrollArea.setWidgetResizable(True)
        self._newSerieWidgetScrollArea.setWidget(self._newSerieWidget)
        self.layout().addWidget(self._newSerieWidgetScrollArea, 0, 3, 4, 2)

        # tomo obj details
        self._tomoObjInfos = TomoObjOverview(self)
        self._tomoObjInfos.setContentsMargins(0, 0, 0, 0)
        self.layout().addWidget(self._tomoObjInfos, 4, 0, 2, 2)

        # connect signals / slot
        self._leftArrowButton.released.connect(self._removeSelectedObjs)
        self._rightArrowButton.released.connect(self._addSelectedObjs)
        self._tomoObjList.selectionModel().selectionChanged.connect(
            self._updateTomoObjInfos
        )

    def selectedTomoObjects(self) -> tuple:
        """
        :return: tuple of tomo object selected on the list
        :rtype: tuple
        """
        items = self._tomoObjList.selectedItems()
        return [item.data(qt.Qt.UserRole) for item in items]

    def _removeSelectedObjs(self, *args, **kwargs):
        for tomo_obj in self.selectedTomoObjects():
            self._newSerieWidget.removeTomoObjToCurrentSerie(tomo_obj)

    def _addSelectedObjs(self, *args, **kwargs):
        for tomo_obj in self.selectedTomoObjects():
            self._newSerieWidget.addTomoObjToCurrentSerie(tomo_obj)

    def _updateTomoObjInfos(self, *args, **kwargs):
        # should
        select_objs = self._tomoObjList.selectedItems()
        if select_objs and len(select_objs) > 0:
            tomo_obj = select_objs[0].data(qt.Qt.UserRole)
            self._tomoObjInfos.setTomoObj(tomo_obj)
        else:
            self._tomoObjInfos.setTomoObj(None)

    # expose API
    def setSerie(self, serie: Serie):
        self._newSerieWidget.setSerie(serie=serie)

    def getSerie(self, *args, **kwargs) -> Serie:
        return self._newSerieWidget.getSerie(*args, **kwargs)

    def addTomoObj(self, tomo_obj):
        self._tomoObjList.add(tomo_obj)

    def addToCurrentSerie(self, tomo_obj):
        self._newSerieWidget.addTomoObjToCurrentSerie(tomo_obj)

    def setSerieName(self, name: str):
        self._newSerieWidget.setSerieName(name=name)


class NewSerieWidget(qt.QWidget):
    sigNameChanged = qt.Signal()
    """Emit when serie name changed"""

    sigUpdated = qt.Signal()
    """
    Emit when the serie has been updated by the tree
    """

    DEFAULT_SERIE_NAME = "my_serie"

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())

        self._nameWidget = qt.QWidget(self)
        self._nameWidget.setLayout(qt.QHBoxLayout())
        self._nameWidget.layout().addWidget(qt.QLabel("serie name", self))
        self._nameQLE = qt.QLineEdit(self.DEFAULT_SERIE_NAME, self)
        self._nameWidget.layout().addWidget(self._nameQLE)
        self.layout().addWidget(self._nameWidget)

        self._serieTree = _SerieDefinitionTree(self, serie_name=self.DEFAULT_SERIE_NAME)
        self.layout().addWidget(self._serieTree)

        # Signal / slot connection
        self._serieTree.itemChanged.connect(self._handleItemUpdate)
        self._serieTree.sigUpdated.connect(self.sigUpdated)
        self._nameQLE.textChanged.connect(self._nameChangedOnQLE)

    def setSerie(self, serie: Serie) -> None:
        with blockSignals(self._nameQLE):
            self._nameQLE.setText(serie.name)
        self._serieTree.setSerie(serie=serie)

    def getSerie(self, *args, **kwargs) -> Serie:
        return self._serieTree.getSerie(*args, **kwargs)

    def _nameChangedOnQLE(self, name):
        with blockSignals(self._serieTree):
            self._serieTree.setSerieName(name)
        self.sigNameChanged.emit()

    def _handleItemUpdate(self, item, column):
        if item == self._serieTree.rootItem:
            old = self.blockSignals(True)
            self._nameQLE.setText(self._serieTree.rootItem.text(0))
            self.blockSignals(old)
        self.sigUpdated.emit()

    def addTomoObjToCurrentSerie(self, tomo_obj: TomwerObject):
        assert isinstance(
            tomo_obj, TomwerObject
        ), f"invalid type {type(tomo_obj)}. {TomwerObject} expected"
        self._serieTree.addTomoObj(tomo_obj)

    def removeTomoObjToCurrentSerie(self, tomo_obj: TomwerObject):
        assert isinstance(
            tomo_obj, TomwerObject
        ), f"invalid type {type(tomo_obj)}. {TomwerObject} expected"
        self._serieTree.removeTomoObj(tomo_obj)

    def getSerieName(self) -> str:
        return self._serieTree.getSerieName()

    def setSerieName(self, name: str):
        self._serieTree.setSerieName(name=name)


class SerieManualDefinitionDialog(qt.QDialog):
    """Dialog to define a serie manually"""

    sigUpdated = qt.Signal()

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())

        self._newSerieWidget = NewSerieWidget(self)

        self._buttons = qt.QDialogButtonBox(parent=self)

        self._addScanButton = qt.QPushButton("Add scan to the serie", parent=self)
        self._buttons.addButton(self._addScanButton, qt.QDialogButtonBox.ActionRole)

        self._removeSelectedButton = qt.QPushButton(
            "Remove selected scans", parent=self
        )
        self._buttons.addButton(
            self._removeSelectedButton, qt.QDialogButtonBox.ActionRole
        )

        self._clearButton = qt.QPushButton("Clear", parent=self)
        self._buttons.addButton(self._clearButton, qt.QDialogButtonBox.ActionRole)

        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self._newSerieWidget.sigNameChanged.connect(self._nameChanged)
        self._newSerieWidget.sigUpdated.connect(self.sigUpdated)
        self._addScanButton.clicked.connect(self.addScanFromFileDialog)
        self._removeSelectedButton.clicked.connect(self.removeSelectedTomoObjs)
        self._clearButton.clicked.connect(self.clearSerie)

    @property
    def n_tomo_objs(self):
        serieTree = self._newSerieWidget._serieTree
        return serieTree.n_tomo_objs

    def _nameChanged(self, new_name):
        serieTree = self._newSerieWidget._serieTree
        with blockSignals(self._serieTree):
            serieTree.setSerieName(name=new_name)
        self.sigUpdated.emit()

    def setSerieName(self, name: str):
        self._newSerieWidget.setSerieName(name=name)

    def getSerieName(self) -> str:
        return self._newSerieWidget.getSerieName()

    def setSerie(self, serie: Serie) -> None:
        self._newSerieWidget.setSerie(serie)

    def getSerie(self, *args, **kwargs) -> Serie:
        return self._newSerieWidget.getSerie(*args, **kwargs)

    def addScanFromFileDialog(self) -> None:
        dialog = QDataDialog(self, multiSelection=True)

        if not dialog.exec_():
            dialog.close()
            return

        foldersSelected = dialog.files_selected()
        for folder in foldersSelected:
            self.addScanFromNxFile(file_=folder, entry=None)

    def addScanFromNxFile(self, file_: str, entry: Optional[str] = None):
        serieTree = self._newSerieWidget._serieTree
        return serieTree.addScanFromNxFile(file_=file_, entry=entry)

    def removeSelectedTomoObjs(self) -> None:
        serieTree = self._newSerieWidget._serieTree
        return serieTree.removeSelectedTomoObjs()

    def getSelectedTomoObjs(self) -> tuple:
        serieTree = self._newSerieWidget._serieTree
        return serieTree.getSelectedTomoObjs()

    def setSelectedTomoObjs(self, scans: Iterable) -> None:
        serieTree = self._newSerieWidget._serieTree
        serieTree.setSelectedTomoObjs(objs=scans)

    def addTomoObj(self, scan: TomwerScanBase) -> None:
        serieTree = self._newSerieWidget._serieTree
        return serieTree.addTomoObj(tomo_obj=scan)

    def removeTomoObj(self, scan: TomwerScanBase) -> None:
        serieTree = self._newSerieWidget._serieTree
        serieTree.removeTomoObj(tomo_obj=scan)

    def clearSerie(self) -> None:
        serieTree = self._newSerieWidget._serieTree
        serieTree.clearTomoObjs()


class SerieAutomaticDefinitionWidget(qt.QWidget):
    pass


class SerieTree(qt.QWidget):
    """
    Widget used to define a scan serie from a list of scans.
    """

    def __init__(self, parent=None, scans=tuple()) -> None:
        self._series = {}
        # associated serie name (key) to tuple (serie, QTreeWidgetItem)
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        self._tree = qt.QTreeWidget(self)
        self._tree.setSelectionMode(qt.QAbstractItemView.MultiSelection)
        self._tree.setColumnCount(2)
        self._tree.setHeaderLabels(("serie", "scan ids"))
        self.layout().addWidget(self._tree)

        # set up
        [self.addSerie(scan) for scan in scans]

    def addSerie(self, serie: Serie):
        if not isinstance(serie, Serie):
            raise TypeError(
                f"{serie} is expected to be an instance of {Serie} not {type(serie)}"
            )
        if serie.name in self._series:
            self.removeSerie(self._series[serie.name][0])

        root_item = qt.QTreeWidgetItem(self._tree)
        root_item.setText(0, serie.name)
        self._series[serie.name] = (serie, root_item)
        for obj in serie:
            scan_item = qt.QTreeWidgetItem(root_item)
            if isinstance(obj, TomwerObject):
                text = obj.get_identifier().to_str()
            else:
                text = obj
            scan_item.setText(1, text)
            scan_item.setFlags(qt.Qt.NoItemFlags)

    def removeSerie(self, serie: Serie):
        if not isinstance(serie, Serie):
            raise TypeError(
                f"{serie} is expected to be an instance of {Serie} not {type(serie)}"
            )
        if serie.name in self._series:
            _, serie_item = self._series.pop(serie.name)
            root = self._tree.invisibleRootItem()
            root.removeChild(serie_item)

    @property
    def n_series(self):
        return len(self._series)

    def series(self) -> tuple:
        series = []
        [series.append(serie) for serie, _ in self._series.values()]
        return tuple(series)

    def clearSelection(self):
        self._tree.selectionModel().clearSelection()

    def setSelectedSeries(self, series):
        self.clearSelection()
        for serie in series:
            serie_item = self._getSerieItem(serie)
            if serie_item is not None:
                serie_item.setSelected(True)

    def _getSerieItem(self, serie: Serie) -> Optional[qt.QTreeWidgetItem]:
        if not isinstance(serie, Serie):
            raise TypeError(
                f"serie is expected to be an instance of {Serie} not {type(serie)}"
            )
        return self._series.get(serie.name, (None, None))[1]

    def getSelectedSeries(self) -> tuple():
        """return selected series"""
        selected = []
        for _, (serie, item) in self._series.items():
            if item.isSelected():
                selected.append(serie)
        return tuple(selected)

    def removeSelected(self) -> None:
        for serie in self.getSelectedSeries():
            self.removeSerie(serie)


class SerieHistoryDialog(qt.QDialog):
    sigSerieSend = qt.Signal(Serie)
    """signal emit when a serie has been selected by the user"""

    sigEditSerie = qt.Signal(Serie)
    """Signal emit when user request to edit a serie"""

    sigHistoryUpdated = qt.Signal()
    """Signal emit when the history has been modified"""

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())

        self._serieList = SerieTree(self)
        self.layout().addWidget(self._serieList)

        self._buttons = qt.QDialogButtonBox(parent=self)

        self._editButton = qt.QPushButton("Edit", parent=self)
        self._buttons.addButton(self._editButton, qt.QDialogButtonBox.ActionRole)

        self._sendButton = qt.QPushButton("Resend", parent=self)
        self._buttons.addButton(self._sendButton, qt.QDialogButtonBox.ActionRole)

        self._removeButton = qt.QPushButton("Remove", parent=self)
        self._buttons.addButton(self._removeButton, qt.QDialogButtonBox.ActionRole)

        self._clearButton = qt.QPushButton("Clear", parent=self)
        self._buttons.addButton(self._clearButton, qt.QDialogButtonBox.ActionRole)

        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self._sendButton.clicked.connect(self.sendSelected)
        self._removeButton.clicked.connect(self.removeSelected)
        self._clearButton.clicked.connect(self.clearSelection)
        self._editButton.clicked.connect(self.editSelected)

    def addSerie(self, serie: Serie):
        old = self.blockSignals(True)
        self._serieList.addSerie(serie)
        self.blockSignals(old)
        self.sigHistoryUpdated.emit()

    def removeSerie(self, serie: Serie):
        old = self.blockSignals(True)
        self._serieList.removeSerie(serie)
        self.blockSignals(old)
        self.sigHistoryUpdated.emit()

    def getSelectedSeries(self):
        return self._serieList.getSelectedSeries()

    def setSelectedSeries(self, series):
        old = self.blockSignals(True)
        self._serieList.setSelectedSeries(series)
        self.blockSignals(old)
        self.sigHistoryUpdated.emit()

    def sendSelected(self):
        for serie in self.getSelectedSeries():
            self.sigSerieSend.emit(serie)

    def editSelected(self):
        selected = self.getSelectedSeries()
        if len(selected) == 0:
            return
        if len(selected) > 1:
            _logger.warning(
                "more than one serie selected for edition. Will only edit the first one"
            )
        self.sigEditSerie.emit(selected[0])

    def removeSelected(self):
        old = self.blockSignals(True)
        self._serieList.removeSelected()
        self.blockSignals(old)
        self.sigHistoryUpdated.emit()

    def clearSelection(self):
        self._serieList.clearSelection()

    def series(self) -> tuple:
        return self._serieList.series()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2021 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "01/12/2016"

import logging

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output, OWBaseWidget
from silx.gui import qt

import tomwer.core.process.control.scanlist
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.control.datalist import GenericScanListWindow

logger = logging.getLogger(__name__)


class DataListOW(OWBaseWidget, openclass=True):
    name = "scan list"
    id = "orange.widgets.tomwer.scanlist"
    description = "List path to reconstructions/scans. Those path will be send to the next box once validated."
    icon = "icons/scanlist.svg"
    priority = 50
    keywords = ["tomography", "file", "tomwer", "folder"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _scanIDs = Setting(list())

    ewokstaskclass = tomwer.core.process.control.scanlist._ScanListPlaceHolder

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    class Inputs:
        data = Input(name="data", type=TomwerScanBase, multiple=True)

    def __init__(self, parent=None):
        """A simple annuary which register all folder containing completed scan

        .. warning: the widget won't check for scan validity and will only
            emit the path to folders to the next widgets

        :param parent: the parent widget
        """
        super().__init__(parent)

        self.widget = GenericScanListWindow(parent=self)
        self.widget.setWindowFlags(qt.Qt.Widget)
        self.widget.sigUpdated.connect(self._updateSettings)
        self._loadSettings()
        layout = gui.vBox(self.mainArea, self.name).layout()
        layout.addWidget(self.widget)
        self.widget._widget._sendButton.clicked.connect(self._sendList)

        # expose API
        self.n_scan = self.widget.n_scan

        # alias used for the 'simple workflow' for now
        self.start = self._sendList
        self.clear = self.widget._widget.clear

    def _sendList(self):
        """Send a signal for each list to the next widget"""
        for data_id, item in self.widget.datalist._myitems.items():
            logger.debug(f"sending {data_id}")
            data_obj = item.data(qt.Qt.UserRole)
            self.Outputs.data.send(data_obj)

    @Inputs.data
    def add(self, scan, *args, **kwargs):
        if scan is None:
            return

        self.widget.add(scan)

    def _loadSettings(self):
        for scan in self._scanIDs:
            assert isinstance(scan, str)
            # kept for backward compatibility since 0.11. To be removed on the future version.
            if "@" in scan:
                entry, file_path = scan.split("@")
                nxtomo_scan = NXtomoScan(entry=entry, scan=file_path)
                self.add(nxtomo_scan)
            else:
                self.add(scan)

    def _updateSettings(self):
        self._scanIDs = []
        for scan in self.widget.datalist._myitems:
            self._scanIDs.append(scan)

    def keyPressEvent(self, event):
        """
        To shortcut orange and make sure the `delete` key will be interpreted we need to overwrite this function
        """
        if event.key() == qt.Qt.Key_Delete:
            self.widget._removeSelected()
        else:
            super().keyPressEvent(event)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = [
    "H. Payno",
]
__license__ = "MIT"
__date__ = "28/09/2020"

import os
import threading
import uuid

from orangewidget import gui, widget
from orangewidget.settings import Setting
from orangewidget.widget import Output
from silx.gui import qt
from silx.gui.utils import concurrent

from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.scanutils import MockBlissAcquisition, MockEDF, MockNXtomo
from tomwer.gui.debugtools.datasetgenerator import DatasetGeneratorDialog


class DatasetGeneratorOW(widget.OWBaseWidget, openclass=True):
    """
    A simple widget to generate on the fly and 'continuously' datasets
    """

    name = "random data generator"
    id = "orangecontrib.tomwer.widgets.debugtools.datasetgeneratorow"
    description = "create on the fly dataset"
    icon = "icons/hammer.png"
    priority = 250
    keywords = ["tomography", "file", "tomwer", "dataset", "debug"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _settings = Setting(dict())

    class Outputs:
        data = Output(name="data", type=object)

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent=parent)
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self.generatorGUI = DatasetGeneratorDialog(parent=self)
        self._layout.addWidget(self.generatorGUI)
        self._displayTimer = qt.QTimer()

        self._loadSettings()

        # connect signal / slot
        self.generatorGUI.sigGenerationStopped.connect(self._stop)
        self.generatorGUI.sigGenerationStarted.connect(self._start)
        self.generatorGUI.sigConfigChanged.connect(self._reset)
        self.generatorGUI.sigCreateOne.connect(self._generate)
        #  connection with the interface to update settings
        self.generatorGUI.sigConfigChanged.connect(self._saveSettings)

    def _start(self):
        self._displayTimer.timeout.connect(self._generateAndRestart)
        self._displayTimer.start(self.generatorGUI.getTimeout())

    def _stop(self):
        self._displayTimer.timeout.disconnect(self._generateAndRestart)
        self._displayTimer.stop()

    def _reset(self):
        if self._displayTimer.isActive():
            self._stop()
            self._start()

    def _generateAndRestart(self):
        self._generate()
        self._displayTimer.start(self.generatorGUI.getTimeout())

    def _generate(self):
        scan_path = os.path.join(
            self.generatorGUI.getRootDir(),
            str(uuid.uuid4()).split("-")[0],
        )
        thread = _DatasetGeneratorThread()
        thread.init(
            type_to_generate=self.generatorGUI.getTypeToGenerate(),
            scan_path=scan_path,
            callback=self.Outputs.data.send,
            n_proj=self.generatorGUI.getNProj(),
            create_ini_dark=self.generatorGUI.isDarkNeededAtBeginning(),
            create_ini_flat=self.generatorGUI.isFlatNeededAtBeginning(),
            dims=self.generatorGUI.getFrameDims(),
        )
        thread.start()
        return thread

    def _loadSettings(self):
        self.generatorGUI.setConfiguration(self._settings)

    def _saveSettings(self):
        self._settings = self.generatorGUI.getConfiguration()


class _DatasetGeneratorThread(threading.Thread):
    def init(
        self,
        type_to_generate,
        scan_path,
        callback,
        n_proj,
        create_ini_dark,
        create_ini_flat,
        dims,
    ):
        self.type_to_generate = type_to_generate
        self.callback = callback
        self.scan_path = scan_path
        self.n_proj = n_proj
        self.create_ini_dark = create_ini_dark
        self.create_ini_flat = create_ini_flat
        self.dims = dims

    def run(self):
        if self.type_to_generate == EDFTomoScan.__name__:
            scan = MockEDF.mockScan(
                scanID=self.scan_path, nRadio=self.n_proj, dim=self.dims[0]
            )
        elif self.type_to_generate == NXtomoScan.__name__:
            scan = MockNXtomo(
                scan_path=self.scan_path,
                n_proj=self.n_proj,
                n_ini_proj=self.n_proj,
                create_ini_dark=self.create_ini_dark,
                create_ini_flat=self.create_ini_flat,
                dim=self.dims[0],
            ).scan
        elif self.type_to_generate == BlissScan.__name__:
            n_darks = 10 if self.create_ini_dark else 0
            n_flats = 10 if self.create_ini_flat else 0
            mock = MockBlissAcquisition(
                n_sample=1,
                n_sequence=1,
                n_scan_per_sequence=1,
                n_darks=n_darks,
                n_flats=n_flats,
                output_dir=self.scan_path,
            )
            scan = mock.create_bliss_scan()
        else:
            raise ValueError(f"type to generate {self.type_to_generate} not recognized")
        concurrent.submitToQtMainThread(self.callback, scan)



import pytest

from tomwer.gui.qfolderdialog import QDataDialog
from tomwer.tests.conftest import qtapp  # noqa F401


@pytest.mark.parametrize("multi_selection", (True, False))
def test_qdata_dialog(
    qtapp,  # noqa F401
    multi_selection,
):
    dialog = QDataDialog(parent=None, multiSelection=multi_selection)
    dialog.files_selected()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "22/20/2020"


import functools

from ewoksorange.bindings import OWEwoksWidgetWithTaskStack
from ewoksorange.bindings.owwidgets import invalid_data
from orangewidget.widget import OWBaseWidget
from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess

from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing


class _SuperviseMixIn(SuperviseProcess):
    def __init__(self, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        self.destroyed.connect(functools.partial(ProcessManager().unregister, self))

    def setCaption(self, caption):
        self.name = caption

    def notify_skip(self, scan, details=None):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.SKIPPED, details=details
        )

    def notify_pending(self, scan, details=None):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.PENDING, details=details
        )

    def notify_succeed(self, scan, details=None):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.SUCCEED, details=details
        )

    def notify_failed(self, scan, details=None):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.FAILED, details=details
        )

    def notify_on_going(self, scan, details=None):
        ProcessManager().notify_dataset_state(
            dataset=scan, process=self, state=DatasetState.ON_GOING, details=details
        )


class SuperviseOW(OWBaseWidget, _SuperviseMixIn, openclass=True):
    """
    A basic OWWidget but registered on the process manager
    """

    want_control_area = False

    def __init__(self, parent, process_id=None):
        OWBaseWidget.__init__(self, parent, process_id=process_id)
        _SuperviseMixIn.__init__(self, process_id=process_id)

    def setCaption(self, caption):
        OWBaseWidget.setCaption(self, caption)
        _SuperviseMixIn.setCaption(self, caption=caption)


class TomwerWithStackStack(
    OWEwoksWidgetWithTaskStack, _SuperviseMixIn, WidgetLongProcessing, openclass=True
):
    def __init__(self, parent, process_id=None, *args, **kwargs):
        OWEwoksWidgetWithTaskStack.__init__(self, parent, args, kwargs)
        _SuperviseMixIn.__init__(self, process_id=process_id)

        self.task_executor_queue.sigComputationStarted.connect(self._startProcessing)
        self.task_executor_queue.sigComputationEnded.connect(self._endProcessing)

    def setCaption(self, caption):
        OWBaseWidget.setCaption(self, caption)
        _SuperviseMixIn.setCaption(self, caption=caption)

    def trigger_downstream(self) -> None:
        # for now ewoksorange send ewoks variable. This will work only if
        # all task are implemented using ewokwidget which is not the case today
        for ewoksname, var in self.get_task_outputs().items():
            channel = self._get_output_signal(ewoksname)
            if invalid_data.is_invalid_data(var.value):
                channel.send(None)  # or channel.invalidate?
            else:
                channel.send(var.value)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


import copy
import logging

import tomoscan.esrf.scan.utils
from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from processview.core.manager import DatasetState, ProcessManager
from silx.gui import qt

import tomwer.core.process.reconstruction.darkref.darkrefscopy
from orangecontrib.tomwer.orange.managedprocess import SuperviseOW
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.reconstruction.darkref.darkrefcopywidget import DarkRefAndCopyWidget
from tomwer.synctools.darkref import QDKRFRP
from tomwer.synctools.stacks.reconstruction.dkrefcopy import DarkRefCopyProcessStack
from tomwer.utils import docstring

from ..utils import WidgetLongProcessing

_logger = logging.getLogger(__name__)


class DarkRefAndCopyOW(SuperviseOW, WidgetLongProcessing):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "reduced darks and flats"
    id = "orange.widgets.tomwer.darkrefs"
    description = (
        "This widget will generate reduced darks and flats for a received scan "
    )
    icon = "icons/darkref.svg"
    priority = 15
    keywords = ["tomography", "dark", "darks", "ref", "refs", "flat", "flats"]

    want_main_area = True
    resizing_enabled = True

    _ewoks_default_inputs = Setting({"data": None, "dark_ref_params": None})

    sigScanReady = qt.Signal(TomwerScanBase)
    """Signal emitted when a scan is ready"""

    ewokstaskclass = (
        tomwer.core.process.reconstruction.darkref.darkrefscopy.DarkRefsCopy
    )

    class Inputs:
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            multiple=True,
        )
        reduced_darks = Input(
            name="reduced dark(s)",
            type=dict,
            doc="dict containing reduced dark(s)",
            multiple=False,
        )
        reduced_flats = Input(
            name="reduced flat(s)",
            type=dict,
            doc="dict of containing reduced flat(s)",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")
        reduced_darks = Output(
            name="reduced dark(s)",
            type=dict,
            doc="DataUrl of the reduced dark. Key is the relative position, value is the reduced dark",
        )
        reduced_flats = Output(
            name="reduced flat(s)",
            type=dict,
            doc="DataUrl of the reduced flat. Key is the relative position, value is the reduced flat",
        )

    def __init__(self, parent=None, reconsparams=None):
        """

        :param QKRFRP reconsparams: reconstruction parameters
        """
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        recons_params = reconsparams or QDKRFRP()
        self._processing_stack = DarkRefCopyProcessStack(process_id=self.process_id)

        dark_ref_params = self._ewoks_default_inputs.get("dark_ref_params", None)
        if dark_ref_params not in ({}, None):
            try:
                recons_params.dkrf.load_from_dict(dark_ref_params)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        self.widget = DarkRefAndCopyWidget(
            save_dir=self._processing_stack._save_dir,
            parent=self,
            reconsparams=recons_params,
            process_id=self.process_id,
        )
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.addWidget(self.widget)
        self.setForceSync = self.widget.setForceSync

        # expose API
        self.setModeAuto = self.widget.set_mode_auto
        self.setRefsFromScan = self.widget.setRefsFromScan
        self.setCopyActive = self.widget.setCopyActive

        # connect signal / slot
        self.widget.sigProcessingStart.connect(self._startProcessing)
        self.widget.sigProcessingEnd.connect(self._endProcessing)
        self.widget.sigScanReady.connect(self.signalReady)
        self.widget.recons_params.sigChanged.connect(self._updateSettingsVals)
        self.widget.sigModeAutoChanged.connect(self._updateSettingsVals)
        self.widget.sigCopyActivationChanged.connect(self._updateSettingsVals)
        self.widget.sigClearCache.connect(self._processing_stack.clear_cache)
        self._processing_stack.sigComputationStarted.connect(self._startProcessing)
        self._processing_stack.sigComputationEnded.connect(self._endProcessing)
        self._processing_stack.sigRefSetted.connect(self.widget.setRefSetBy)

        # load some other copy parameters
        if dark_ref_params not in ({}, None):
            try:
                if "activate" in dark_ref_params:
                    self.widget.setCopyActive(dark_ref_params.pop("activate"))
                if "auto" in dark_ref_params:
                    auto_mode = dark_ref_params.pop("auto")
                    # insure backward compatibility. Has beem saved as a tuple (this was a typo)
                    if isinstance(auto_mode, tuple) and isinstance(auto_mode[0], bool):
                        auto_mode = auto_mode[0]
                    self.widget.setModeAuto(auto_mode)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        self.setCaption(self.windowTitle())

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    @Inputs.data
    def process(self, scanID, *args, **kwargs):
        if scanID is None:
            return
        assert isinstance(scanID, TomwerScanBase)
        ProcessManager().notify_dataset_state(
            dataset=scanID, process=self, state=DatasetState.PENDING
        )
        configuration = self.widget.recons_params.to_dict()
        configuration["process_only_copy"] = False  # we never only do copy from the gui
        configuration["process_only_dkrf"] = not self.widget.isCopyActive()
        configuration["mode_auto"] = self.widget.isOnModeAuto()

        self._processing_stack.add(copy.copy(scanID), configuration=configuration)

    @Inputs.reduced_darks
    def received_reduced_darks(self, reduced_darks: dict):
        # we consider if the darks are provided then the user want to use those and don't copy them manually
        self.setModeAuto(False)
        reduced_darks.pop("reduce_frames_name", None)
        self.setReducedDarks(reduced_darks)

    @Inputs.reduced_flats
    def received_reduced_flats(self, reduced_flats: dict):
        # we consider if the darks are provided then the user want to use those and don't copy them manually
        self.setModeAuto(False)
        reduced_flats.pop("reduce_frames_name", None)
        self.setReducedFlats(reduced_flats)

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        self.process(dataset)

    def signalReady(self, scanID):
        assert isinstance(scanID, TomwerScanBase)
        self.Outputs.data.send(scanID)
        self.sigScanReady.emit(scanID)

    def _updateSettingsVals(self):
        self._ewoks_default_inputs = {
            "data": None,
            "dark_ref_params": self.widget.recons_params.to_dict(),
        }
        self._ewoks_default_inputs["dark_ref_params"][
            "auto"
        ] = self.widget.isOnModeAuto()
        self._ewoks_default_inputs["dark_ref_params"][
            "activate"
        ] = self.widget.isCopyActive()

    @property
    def recons_params(self):
        return self.widget.recons_params

    def close(self):
        self.widget.close()
        super(DarkRefAndCopyOW, self).close()

    def _endProcessing(self, scan):
        WidgetLongProcessing._endProcessing(self, scan)
        self.Outputs.data.send(scan)
        self.sigScanReady.emit(scan)
        if scan.reduced_darks not in (None, {}):
            # we want to send those in relative position to have something generic. This is a convention for now
            reduced_darks = scan.reduced_darks
            reduced_darks.pop("reduce_frames_name", None)
            self.Outputs.reduced_darks.send(
                tomoscan.esrf.scan.utils.from_absolute_reduced_frames_to_relative(
                    reduced_frames=reduced_darks, scan=scan
                )
            )
        if scan.reduced_flats not in (None, {}):
            # we want to send those in relative position to have something generic. This is a convention for now
            reduced_flats = scan.reduced_flats
            reduced_flats.pop("reduce_frames_name", None)
            self.Outputs.reduced_flats.send(
                tomoscan.esrf.scan.utils.from_absolute_reduced_frames_to_relative(
                    reduced_frames=reduced_flats, scan=scan
                )
            )
        _logger.info(f"{scan} ended")

    def setReducedDarks(self, darks: dict):
        self.widget._refCopyWidget.save_darks_to_be_copied(darks)

    def setReducedFlats(self, flats: dict):
        self.widget._refCopyWidget.save_flats_to_be_copied(flats)



from silx.gui import qt


Y_AXIS_DOWNWARD = True
"""Default axis direction Y"""

EDITING_BACKGROUND_COLOR = qt.QColor("#c3d0d6")
"""Background color used to notify the user that application waits end of edition before taking it into account"""



# from typing import Optional
# from orangewidget import gui
# from orangewidget.settings import Setting

# from silx.gui import qt
# import weakref
# from ewoksorange.bindings.owwidgets import OWEwoksWidgetOneThreadPerRun

# from tomwer.core.scan.scanbase import TomwerScanBase
# import tomwer.core.process.icat.gallery
# from tomwer.gui.icat.gallery import GalleryWidget, _GalleryOutputDir
# from processview.core.superviseprocess import SuperviseProcess


# class SaveToGalleryAndPublishOW(
#     OWEwoksWidgetOneThreadPerRun,
#     SuperviseProcess,
#     ewokstaskclass=tomwer.core.process.icat.gallery.SaveToGalleryAndPublishTask,
# ):
#     """
#     This widget can receive 'screenshots' signal (expected as a tuple (source, screenshots_dict)
#     source is the tomo obj associated (so we can automate deduction of the output dir)
#     screenshots_dict is a dict of str: DataUrl
#     """

#     name = "Save to gallery and publish"
#     id = "orangecontrib.widgets.tomwer.icat.MoveToGalleryOW.SaveToGalleryAndPublishOW"
#     description = "This widget will save screenshot of reconstruction to a specific location (acquisition gallery as the default end point) \n"
#     icon = "icons/add_gallery.svg"
#     priority = 63
#     keywords = [
#         "tomography",
#         "tomwer",
#         "tomo_obj",
#         "screenshot",
#         "acquisition",
#         "gallery",
#     ]

#     want_main_area = True
#     want_control_area = False
#     resizing_enabled = True

#     _ewoks_default_inputs = Setting(
#         {
#             "format": "png",
#             "output_location_mode": "default",
#             "custom_output": None,
#             "screenshots": None,
#             "overwrite": False,
#         }
#     )

#     _ewoks_inputs_to_hide_from_orange = (
#         "proposal",
#         "dataset",
#         "beamline",
#         "dry_run",
#         "__process__",
#     )

#     def __init__(self, *args, **kwargs):
#         super().__init__(*args, **kwargs)
#         SuperviseProcess.__init__(self)
#         self._scan = None
#         self._widget = GalleryWidget(parent=self)

#         self._box = gui.vBox(self.mainArea, self.name)
#         layout = self._box.layout()
#         layout.addWidget(self._widget)

#         # load settings
#         self._widget.setConfiguration(self._ewoks_default_inputs)

#         # connect signal / slot
#         self._widget.sigConfigChanged.connect(self._updateSettings)

#     def _updateSettings(self):
#         self._ewoks_default_inputs = self._widget.getConfiguration()

#     def setScan(self, scan: TomwerScanBase):
#         self._scan = weakref.ref(scan)
#         self._widget._publishConfig.setScan(scan=scan)

#     def getScan(self) -> Optional[TomwerScanBase]:
#         if self._scan is None:
#             return None
#         else:
#             return self._scan()

#     def handleNewSignals(self) -> None:
#         """Invoked by the workflow signal propagation manager after all
#         signals handlers have been called.
#         """
#         # update the widget when receive the scan (proposal, dataset...)
#         screenshots = self.get_task_input_value("screenshots", None)
#         if screenshots is None:
#             return
#         else:
#             scan = screenshots.scan
#             if scan != self.getScan():
#                 self.setScan(scan)

#         if self.get_task_input_value("screenshots", None) is None:
#             return
#         else:
#             super().handleNewSignals()

#     def get_task_inputs(self):
#         task_inputs = super().get_task_inputs()
#         configuration = self._widget.getConfiguration()

#         use_custom_output_dir = _GalleryOutputDir.OutputDirMode.from_value(
#             configuration.get("output_location_mode", "default")
#         )
#         # if user requested to overwrite the output directory
#         if use_custom_output_dir is _GalleryOutputDir.OutputDirMode.PROPOSAL_GALLERY:
#             task_inputs["screenshots"]["output_dir"] = configuration["custom_output"]

#         task_inputs["overwrite"] = configuration["overwrite"]
#         task_inputs["format"] = configuration["output_format"]
#         task_inputs["beamline"] = configuration["beamline"]
#         task_inputs["dataset"] = configuration["dataset"]
#         task_inputs["proposal"] = configuration["proposal"]
#         task_inputs["__process__"] = weakref.ref(self)
#         return task_inputs

#     def sizeHint(self):
#         return qt.QSize(500, 200)



from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.gui.control.emailnotifier import Emailwidget


def test_Emailwidget(qtapp):  # noqa F811
    """test of Emailwidget"""
    widget = Emailwidget()
    widget.show()

    assert widget.getConfiguration() == {
        "from_addr": "",
        "to_addrs": ("",),
        "subject": "tomwer: {tomo_obj_short_id} received",
        "text": "{tomo_obj_id} received at {timestamp} \n\nprocesses:\n{dataset_processing_states}\n\nfile listening:\n{ls_tomo_obj}\n\nInfo: {footnote}",
        "port": 0,
        "host": "smtps.esrf.fr",
    }

    new_configuration = {
        "from_addr": "toto.esrf.fr",
        "to_addrs": "toto.esrf.fr;tata.esrf.fr",
        "subject": "this is a scan",
        "text": "",
        "port": 445,
        "host": "smtps.esrf.en",
    }
    widget.setConfiguration(new_configuration)

    configuration = widget.getConfiguration()
    new_configuration.pop("to_addrs")
    assert configuration.pop("to_addrs") in (
        ("toto.esrf.fr", "tata.esrf.fr"),
        ("tata.esrf.fr", "toto.esrf.fr"),
    )
    assert configuration == new_configuration



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

try:
    from orangecanvas.utils.pkgmeta import get_distribution
except ImportError:
    from pkg_resources import get_distribution

ICON = "../widgets/icons/tomwer.png"

BACKGROUND = "#C0CCFF"


# Entry point for main Orange categories/widgets discovery
def widget_discovery(discovery):
    dist = get_distribution("tomwer")
    pkgs = [
        "orangecontrib.tomwer.widgets.cluster",
        "orangecontrib.tomwer.widgets.control",
        "orangecontrib.tomwer.widgets.debugtools",
        "orangecontrib.tomwer.widgets.edit",
        # "orangecontrib.tomwer.widgets.icat",
        "orangecontrib.tomwer.widgets.reconstruction",
        # "orangecontrib.tomwer.widgets.stitching",
        "orangecontrib.tomwer.widgets.visualization",
        "orangecontrib.tomwer.widgets.other",
    ]
    for pkg in pkgs:
        discovery.process_category_package(pkg, distribution=dist)


WIDGET_HELP_PATH = (
    # Used for development.
    # You still need to build help pages using
    # make htmlhelp
    # inside doc folder
    # (
    #     "/home/payno/Documents/dev/tomography/tomwer/build/html/canvas/widgets/widgets.html",
    #     None,
    # ),
    ("https://tomwer.readthedocs.io/en/latest/canvas/widgets/widgets.html", None),
)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "03/08/2020"


import code
import logging

import tomwer.version
from tomwer.core.process.task import Task
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.scanutils import data_identifier_to_scan
from tomwer.core.utils.volumeutils import volume_identifier_to_volume

_logger = logging.getLogger(__name__)


class PythonScript(
    Task,
    optional_input_names=("data", "volume", "serialize_output_data"),
    output_names=("data", "volume"),
):
    def run(self):
        # load data
        scan = data_identifier_to_scan(self.inputs.data)
        if isinstance(scan, dict):
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        elif isinstance(scan, (TomwerScanBase, type(None))):
            scan = scan
        else:
            raise ValueError(f"input type of {scan}: {type(scan)} is not managed")
        # load volume
        volume = volume_identifier_to_volume(self.inputs.volume)

        cfg = self.get_configuration()
        interpreter = code.InteractiveConsole(locals={"in_data": scan})
        interpreter = code.InteractiveConsole(locals={"in_volume": volume})
        interpreter.runcode(cfg["scriptText"])
        out_data = data_identifier_to_scan(interpreter.locals.get("out_data"))
        out_volume = data_identifier_to_scan(interpreter.locals.get("out_volume"))

        # register process
        if isinstance(scan, EDFTomoScan):
            entry = None
        else:
            entry = scan.entry

        try:
            self.register_process(
                process_file=scan.process_file,
                entry=entry,
                configuration={"scriptText": self.get_configuration()["scriptText"]},
                results=None,
                process_index=scan.pop_process_index(),
                overwrite=True,
            )
        except Exception as e:
            _logger.error("Fail to register process. Error is " + str(e))

        if out_data is not None and self.get_input_value("serialize_output_data", True):
            self.outputs.data = out_data.to_dict()
        else:
            self.outputs.data = out_data
        self.outputs.volume = out_volume

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "Python script"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        return "Execute some random python code"



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "18/02/2018"


import functools

from silx.gui import qt

from tomwer.core.process.control.datawatcher import status
from tomwer.core.process.control.datawatcher.datawatcher import _DataWatcher


class _DWConfigurationWidget(qt.QWidget):
    startByOldestStateChanged = qt.Signal(bool)

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._observationMethod = _ObservationMethodSelector(parent=self)
        self.layout().addWidget(self._observationMethod)

        # checkbox start scan from oldest
        self._qcboldest = qt.QCheckBox("Start scan by the oldest", parent=self)
        tooltip = (
            "If NOT activated will explore folders from the latest to "
            "the newest. Otherwise will explore the folders from the "
            "newest to the oldest."
        )
        self._qcboldest.setToolTip(tooltip)
        self.startByOldestStateChanged = self._qcboldest.toggled
        self.layout().addWidget(self._qcboldest)

    def getMode(self) -> tuple:
        return self._observationMethod.getMode()

    def setMode(self, mode: tuple):
        self._observationMethod.setMode(mode)


class _ObservationMethodSelector(qt.QWidget):
    sigSelectionChanged = qt.Signal(tuple)
    """Return the selection made as a string and some information if needed in
    a dictionary"""

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.__button_group = qt.QButtonGroup(self)

        self.setLayout(qt.QVBoxLayout())
        self._edfObservations = _SpecEDFObservationMethodSelector(self)
        self.layout().addWidget(self._edfObservations)
        self._blissScanObservations = _BlissHDF5ObservationMethodSelector(self)
        self.layout().addWidget(self._blissScanObservations)
        self._nxtomoObservations = _NXtomoObservationMethodSelector(self)
        self.layout().addWidget(self._nxtomoObservations)

        self._edfObservations.sigSelectionChanged.connect(
            self._propagateSigSelectionChanged
        )

        self.__button_group.addButton(self._edfObservations._qrbInfo)
        self.__button_group.addButton(self._edfObservations._qrbUserEntry)
        self.__button_group.addButton(self._edfObservations._qrbXml)
        self.__button_group.addButton(self._blissScanObservations._qrbBlissHDF5)
        self.__button_group.addButton(self._nxtomoObservations._qrbNXtomo)

        # connect signal / slot
        self._edfObservations.sigSelectionChanged.connect(
            self._propagateSigSelectionChanged
        )
        self._blissScanObservations._qrbBlissHDF5.toggled.connect(
            functools.partial(
                self._propagateSigSelectionChanged, (status.BLISS_SCAN_END,)
            )
        )
        self._nxtomoObservations._qrbNXtomo.toggled.connect(
            functools.partial(self._propagateSigSelectionChanged, (status.NXtomo_END,))
        )

    def _propagateSigSelectionChanged(self, infos: tuple):
        self.sigSelectionChanged.emit(infos)

    def getMode(self):
        if self._blissScanObservations._qrbBlissHDF5.isChecked():
            return (status.BLISS_SCAN_END,)
        elif self._nxtomoObservations._qrbNXtomo.isChecked():
            return (status.NXtomo_END,)
        else:
            return self._edfObservations.getMode()

    def setMode(self, mode: tuple):
        if mode[0] == status.BLISS_SCAN_END:
            self._blissScanObservations._qrbBlissHDF5.setChecked(True)
        elif mode[0] == status.NXtomo_END:
            self._nxtomoObservations._qrbNXtomo.setChecked(True)
        else:
            self._edfObservations.setMode(mode)


class _NXtomoObservationMethodSelector(qt.QGroupBox):
    def __init__(self, parent):
        super().__init__(parent, title="End acquisition observation method for NXtomo")
        self.setLayout(qt.QVBoxLayout())

        self._qrbNXtomo = qt.QRadioButton(status.NXtomo_END, parent=self)
        self.layout().addWidget(self._qrbNXtomo)


class _BlissHDF5ObservationMethodSelector(qt.QGroupBox):
    def __init__(self, parent):
        super().__init__(
            parent, title="End acquisition observation method for bliss-hdf5"
        )
        self.setLayout(qt.QVBoxLayout())

        self._qrbBlissHDF5 = qt.QRadioButton(status.BLISS_SCAN_END, parent=self)
        self.layout().addWidget(self._qrbBlissHDF5)


class _SpecEDFObservationMethodSelector(qt.QGroupBox):
    """Group box allowing selection of an observation method"""

    sigSelectionChanged = qt.Signal(tuple)
    """Return the selection made as a string and some information if needed in
    a dictionary"""

    def __init__(self, parent):
        super().__init__(parent, title="End acquisition observation method for edf")
        self.setLayout(qt.QVBoxLayout())

        self._qrbXml = qt.QRadioButton(status.DET_END_XML, parent=self)
        self.layout().addWidget(self._qrbXml)
        self._qrbInfo = qt.QRadioButton(status.PARSE_INFO_FILE, parent=self)
        self.layout().addWidget(self._qrbInfo)
        self._qrbAnyEDF = qt.QRadioButton(status.EDF_SCAN_END, parent=self)
        self.layout().addWidget(self._qrbAnyEDF)

        self._qwUserEntry = qt.QWidget(parent=self)
        self.layout().addWidget(self._qwUserEntry)
        self._qrbUserEntry = qt.QRadioButton(status.DET_END_USER_ENTRY, parent=self)
        self.layout().addWidget(self._qrbUserEntry)

        widgetFilePtrn = qt.QWidget(parent=self)
        widgetFilePtrn.setLayout(qt.QHBoxLayout())

        widgetFilePtrn.layout().addWidget(
            qt.QLabel(text="pattern: ", parent=widgetFilePtrn)
        )
        self._qleFilePattern = qt.QLineEdit(text="", parent=self._qwUserEntry)
        widgetFilePtrn.layout().addWidget(self._qleFilePattern)
        self.widgetFilePtrn = widgetFilePtrn
        self.layout().addWidget(self.widgetFilePtrn)

        self._qrbXml.setChecked(_DataWatcher.DEFAULT_OBS_METH == status.DET_END_XML)
        self._qrbInfo.setChecked(
            _DataWatcher.DEFAULT_OBS_METH == status.PARSE_INFO_FILE
        )
        self._qrbUserEntry.setChecked(
            _DataWatcher.DEFAULT_OBS_METH == status.DET_END_USER_ENTRY
        )

        self.widgetFilePtrn.setVisible(self._qrbUserEntry.isVisible())

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer)

        # set up
        self._qrbAnyEDF.hide()  # by default for the DataWatcher compatibility this one is hidded

        # deal with connections
        self._qrbUserEntry.toggled.connect(self.widgetFilePtrn.setVisible)
        self._qrbXml.toggled.connect(self._selectionChanged)
        self._qrbInfo.toggled.connect(self._selectionChanged)
        self._qrbAnyEDF.toggled.connect(self._selectionChanged)
        self._qrbUserEntry.toggled.connect(self._selectionChanged)
        self._qleFilePattern.editingFinished.connect(self._selectionChanged)

        # add some tooltips
        t = """If a file with this pattern is found in the [scan] folder then
            we will consider the acquisition as ended"""
        self._qrbUserEntry.setToolTip(t)

        t = """Wild charracter allowed"""
        self._qleFilePattern.setToolTip(t)

        t = """If we founf the [scan].xml in the [scan] folder then
            we will consider the acquisition ended"""
        self._qrbXml.setToolTip(t)

        t = """We will look for the [scan].info file in the [scan]
            directory. If it exists then we will parse it to get the number of
            .edf file we should have and wait for all of them to be acquired
            (also checking file size)"""
        self._qrbInfo.setToolTip(t)

    def _selectionChanged(self):
        mode = self.getMode()
        if mode is not None:
            self.sigSelectionChanged.emit((mode,))

    def getMode(self):
        if self._qrbXml.isChecked():
            return (status.DET_END_XML,)
        elif self._qrbInfo.isChecked():
            return (status.PARSE_INFO_FILE,)
        elif self._qrbAnyEDF.isChecked():
            return (status.EDF_SCAN_END,)
        elif self._qrbUserEntry.isChecked():
            return (status.DET_END_USER_ENTRY, {"pattern": self._qleFilePattern.text()})
        else:
            return None

    def setMode(self, mode: tuple):
        if mode[0] == status.DET_END_XML:
            self._qrbXml.setChecked(True)
        elif mode[0] == status.PARSE_INFO_FILE:
            self._qrbInfo.setChecked(True)
        elif mode[0] == status.EDF_SCAN_END:
            self._qrbAnyEDF.setChecked(True)
        elif mode[0] == status.DET_END_USER_ENTRY:
            self._qrbUserEntry.setChecked(True)
            if mode[1] not in (None, ""):
                self._qleFilePattern.setText(mode[1])



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2020 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "11/02/2020"


from typing import Union

from tomwer.core.process.reconstruction.nabu.utils import _NabuStages
from tomwer.gui.configuration.level import ConfigurationLevel


class _FilteringObject:
    """
    Simple class to define if the widget should be visible or not if visible
    (at some moment the 'option' wudget should be visible) and if filtered.
    Should avoid some conflict when set set visible
    """

    def __init__(self, widget):
        self._widget = widget
        self._visible = True
        # is this
        self._filtered = False

    def setVisible(self, visible):
        self._visible = visible
        self._updateVisibility()

    def setFiltered(self, filtered):
        self._filtered = filtered
        self._updateVisibility()

    def _updateVisibility(self):
        self._widget.setVisible(self._visible and self._filtered)


class _NabuStageConfigBase:
    """Define interface for a specific nabu stage configuration widget
    (or a part of the stage configuration)
    """

    def __init__(self, stage: Union[_NabuStages, str]):
        self.__stage = _NabuStages.from_value(stage)
        self._registeredWidgets = {}
        # list required widgets. Key is widget, value is the configuration
        # level

    def registerWidget(self, widget, config_level):
        """register a widget with a configuration level.

        :returns: _FilteringObject to use to define widget visibility
        """
        filteringObj = _FilteringObject(widget=widget)
        self._registeredWidgets[filteringObj] = ConfigurationLevel.from_value(
            config_level
        )
        return filteringObj

    def getConfiguration(self) -> dict:
        raise NotImplementedError("Base class")

    def setConfiguration(self, config) -> None:
        raise NotImplementedError("Base class")

    def getStage(self) -> _NabuStages:
        return self.__stage

    def setConfigurationLevel(self, level):
        for widget in self._registeredWidgets:
            filtered = self._registeredWidgets[widget] <= level
            widget.setFiltered(filtered)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "23/03/2021"


from silx.gui import qt
from functools import partial

from tomwer.gui import icons as tomwer_icons
from tomwer.gui.control.tomoobjdisplaymode import DisplayMode


class NXTomomillParamsAction(qt.QAction):
    """
    Action to display a window with nxtomomill configuration
    """

    def __init__(self, parent):
        icon = tomwer_icons.getQIcon("parameters")

        qt.QAction.__init__(self, icon, "filter configuration", parent)
        self.setToolTip("Open dialog to configure nxtomomill parameters")
        self.setCheckable(False)


class CFGFileActiveLabel(qt.QLabel):
    """Label used to display if the .cfg file is active or not"""

    def __init__(self, parent):
        super().__init__(parent)
        icon = tomwer_icons.getQIcon("cfg_file_inactive")
        self.setToolTip("no valid cfg file provided")
        self.setPixmap(icon.pixmap(self.width(), self.height()))

    def setActive(self, active=True):
        if active is True:
            icon = tomwer_icons.getQIcon("cfg_file_active")
            tooltip = "will use the provided .cfg file"
        else:
            icon = tomwer_icons.getQIcon("cfg_file_inactive")
            tooltip = "will use default configuration"

        self.setPixmap(icon.pixmap(self.width(), self.height()))
        self.setToolTip(tooltip)

    def setInactive(self):
        self.setActive(active=False)


class TomoObjDisplayModeToolButton(qt.QToolButton):
    """
    Button to change the way tomo object are displayed.
    Either using the full url or only a 'short' description.
    """

    sigDisplayModeChanged = qt.Signal(str)

    _SHORT_DESC_TOOLTIP = "Use a short description of the tomo object. Two different scans can have the same short desciption"
    _URL_TOOLTIP = (
        "Use the full url to display the tomo object. Url is guaranted to be unique."
    )

    def __init__(self, parent=None) -> None:
        super().__init__(parent)

        self._shortDescIcon = tomwer_icons.getQIcon("short_description")
        shortDescAction = qt.QAction(self._shortDescIcon, "short description", self)
        shortDescAction.setToolTip(self._SHORT_DESC_TOOLTIP)

        self._urlIcon = tomwer_icons.getQIcon("url")
        urlDescAction = qt.QAction(self._urlIcon, "url", self)
        urlDescAction.setToolTip(self._URL_TOOLTIP)

        menu = qt.QMenu(self)
        menu.addAction(shortDescAction)
        menu.addAction(urlDescAction)
        self.setMenu(menu)
        self.setPopupMode(qt.QToolButton.InstantPopup)

        # set up
        self.setDisplayMode(DisplayMode.SHORT)

        # connect signal / slot
        shortDescAction.triggered.connect(
            partial(self.setDisplayMode, DisplayMode.SHORT)
        )
        urlDescAction.triggered.connect(partial(self.setDisplayMode, DisplayMode.URL))

    def setDisplayMode(self, mode: DisplayMode):
        mode = DisplayMode.from_value(mode)
        if mode is DisplayMode.SHORT:
            self.setIcon(self._shortDescIcon)
            self.setToolTip(self._SHORT_DESC_TOOLTIP)
        elif mode is DisplayMode.URL:
            self.setIcon(self._urlIcon)
            self.setToolTip(self._URL_TOOLTIP)
        else:
            raise ValueError(f"display mode {mode} not handled")
        self.sigDisplayModeChanged.emit(mode.value)



#!/usr/bin/env python
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module describe tomwer applications which are available  through
the silx launcher.

Your environment should provide a command `tomwer`. You can reach help with
`tomwer --help`, and check the version with `tomwer --version`.
"""

__authors__ = ["V. Valls", "P. Knobel", "H. Payno"]
__license__ = "MIT"
__date__ = "10/11/2020"


import sys
from collections import namedtuple
import tomwer.version
from tomwer.utils import Launcher as _Launcher
from tomwer.core.utils.deprecation import deprecated_warning

DeprecationWarning = namedtuple(
    "DeprecationWarning", ["since", "reason", "replacement"]
)


class Launcher(_Launcher):
    """
    Manage launch of module.

    Provides an API to describe available commands and feature to display help
    and execute the commands.
    """

    def __init__(
        self, prog=None, usage=None, description=None, epilog=None, version=None
    ):
        super().__init__(
            prog=prog,
            usage=usage,
            description=description,
            epilog=epilog,
            version=version,
        )
        self._deprecations = {}
        "deprecations with prog names as key and deprecation info as values"

    def add_command(
        self,
        name=None,
        module_name=None,
        description=None,
        command=None,
        deprecated=False,
        deprecated_since_version=None,
        deprecated_reason=None,
        deprecated_replacement=None,
    ):
        super().add_command(
            name=name, module_name=module_name, description=description, command=command
        )
        if deprecated:
            self._deprecations[name] = DeprecationWarning(
                since=deprecated_since_version,
                reason=deprecated_reason,
                replacement=deprecated_replacement,
            )

    def execute(self, argv=None):
        if argv is None:
            argv = sys.argv

        if len(argv) <= 1:
            self.print_help()
            return 0

        command_name = argv[1]

        if command_name in self._deprecations:
            deprecation_info = self._deprecations[command_name]
            deprecated_warning(
                type_="application",
                name=command_name,
                reason=deprecation_info.reason,
                replacement=deprecation_info.replacement,
                since_version=deprecation_info.since,
            )
        super().execute(argv=argv)


def main():
    """Main function of the launcher

    This function is referenced in the setup.py file, to create a
    launcher script generated by setuptools.

    :rtype: int
    :returns: The execution status
    """
    _version = tomwer.version.version
    launcher = Launcher(prog="tomwer", version=_version)
    launcher.add_command(
        "axis",
        module_name="tomwer.app.axis",
        description="Compute center of rotation of a scan or "
        "between two projections",
    )
    launcher.add_command(
        "canvas", module_name="tomwer.app.canvas", description="open the orange-canvas"
    )
    launcher.add_command(
        "dark-ref",
        module_name="tomwer.app.reducedarkflat",
        description="Compute mean / median darks and flat fields",
        deprecated=True,
        deprecated_reason="improve name",
        deprecated_replacement="reduce-dark-flat",
        deprecated_since_version=1.2,
    )
    launcher.add_command(
        "dark-ref-patch",
        module_name="tomwer.app.patchrawdarkflat",
        description="patch raw dark and / or flat to an NXTomo entry",
        deprecated=True,
        deprecated_reason="improve name",
        deprecated_replacement="patch-raw-dark-flat",
        deprecated_since_version=1.2,
    )
    launcher.add_command(
        "diff-frame",
        module_name="tomwer.app.diffframe",
        description="Tool to compare any frame from the acquisition",
    )
    launcher.add_command(
        "image-key-editor",
        module_name="tomwer.app.imagekeyeditor",
        description="Tool to modify in place frames image key parameter.",
    )
    launcher.add_command(
        "image-key-upgrader",
        module_name="tomwer.app.imagekeyupgrader",
        description="Tool to upgrade in place frames image key parameter.",
    )
    launcher.add_command(
        "norm-intensity",
        module_name="tomwer.app.intensitynormalization",
        description="Tool to compute intensity normalization.",
    )
    launcher.add_command(
        "multi-cor",
        module_name="tomwer.app.multicor",
        description="Compute a slice with different values of cor",
    )
    launcher.add_command(
        "multi-pag",
        module_name="tomwer.app.multipag",
        description="compute a slice with different values of delta / beta values",
    )
    launcher.add_command(
        "nabu",
        module_name="tomwer.app.nabuapp",
        description="Use nabu to reconstruct a given scan",
    )
    launcher.add_command(
        "nxtomo-editor",
        module_name="tomwer.app.nxtomoeditor",
        description="Use to edit an NXtomo",
    )
    launcher.add_command(
        "ows-to-script",
        module_name="tomwer.app.ows_to_script",
        description="Convert a .ows file to python script",
    )
    launcher.add_command(
        "patch-raw-dark-flat",
        module_name="tomwer.app.patchrawdarkflat",
        description="patch raw dark and / or flat to an NXTomo entry",
    )
    launcher.add_command(
        "reduce-dark-flat",
        module_name="tomwer.app.reducedarkflat",
        description="Compute reduce darks / flats (mean / median of raw darks and flats)",
    )
    launcher.add_command(
        "rsync",
        module_name="tomwer.app.rsync",
        description="Tool for synchronization between "
        "acquisition contained in a folder and "
        "a target folder",
    )
    launcher.add_command(
        "sample-moved",
        module_name="tomwer.app.samplemoved",
        description="Interface to check if a sample have " "moved during acquisition ",
    )
    launcher.add_command(
        "slice-stack",
        module_name="tomwer.app.slicestack",
        description="Display each reconstruction contained "
        "in the ROOT_DIR or sub-folders",
    )
    launcher.add_command(
        "sinogram-viewer",
        module_name="tomwer.app.sinogramviewer",
        description="Allows to compute on the fly sinogram " "and display them",
    )
    launcher.add_command(
        "scan-viewer",
        module_name="tomwer.app.scanviewer",
        description="Scan viewer / browser",
    )
    launcher.add_command(
        "stop-data-listener",
        module_name="tomwer.app.stopdatalistener",
        description="Send SIGTERM or SIGKILL signal to the process occupying"
        "rpc-server port",
    )
    launcher.add_command(
        "test", module_name="tomwer.app.test_", description="Launch tomwer unittest"
    )
    launcher.add_command(
        "z-stitching",
        module_name="tomwer.app.zstitching",
        description="Launch z-stitcher interface",
    )
    status = launcher.execute(sys.argv)
    return status


if __name__ == "__main__":
    # executed when using python -m PROJECT_NAME
    status = main()
    sys.exit(status)
